# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.
import platform
# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


#  Standard Library Imports
from multiprocessing import Manager
import tkinter
from tkinter import messagebox
import multiprocessing as mp
import threading
import sys
import os
import time

# Third Party Imports

# Local View Imports
from navigate.view.main_application_window import MainApp as view
from navigate.view.popups.camera_view_popup_window import CameraViewPopupWindow
from navigate.view.popups.feature_list_popup import FeatureListPopup

# Local Sub-Controller Imports
from navigate.controller.configuration_controller import ConfigurationController
from navigate.controller.sub_controllers import (
    KeystrokeController,
    WaveformTabController,
    StageController,
    CameraSettingController,
    CameraViewController,
    MultiPositionController,
    ChannelsTabController,
    AcquireBarController,
    FeaturePopupController,
    MenuController,
    PluginsController,
    # MicroscopePopupController,
    # AdaptiveOpticsPopupController,
)

from navigate.controller.thread_pool import SynchronizedThreadPool

# Local Model Imports
from navigate.model.model import Model
from navigate.model.concurrency.concurrency_tools import ObjectInSubprocess

# Misc. Local Imports
from navigate.config.config import (
    load_configs,
    update_config_dict,
    verify_experiment_config,
    verify_waveform_constants,
    get_navigate_path,
)
from navigate.tools.file_functions import create_save_path, save_yaml_file
from navigate.tools.common_dict_tools import update_stage_dict
from navigate.tools.multipos_table_tools import update_table
from navigate.tools.common_functions import combine_funcs

# Logger Setup
import logging

p = __name__.split(".")[1]
logger = logging.getLogger(p)


class Controller:
    """Navigate Controller"""

    def __init__(
        self,
        root,
        splash_screen,
        configuration_path,
        experiment_path,
        waveform_constants_path,
        rest_api_path,
        waveform_templates_path,
        args,
    ):
        """Initialize the Navigate Controller.

        Parameters
        ----------
        root : Tk top-level widget.
            Tk.tk GUI instance.
        splash_screen : Tk top-level widget.
            Tk.tk GUI instance.
        configuration_path : string
            Path to the configuration yaml file.
            Provides global microscope configuration parameters.
        experiment_path : string
            Path to the experiment yaml file.
            Provides experiment-specific microscope configuration.
        waveform_constants_path : string
            Path to the waveform constants yaml file.
            Provides magnification and wavelength-specific parameters.
        rest_api_path : string
            Path to the REST API yaml file.
            Provides REST API configuration parameters.
        waveform_templates_path : string
            Path to the waveform templates yaml file.
            Provides waveform templates for each channel.
        *args :
            Command line input arguments for non-default
            file paths or using synthetic hardware modes.
        """

        #: Object: Thread pool for the controller.
        self.threads_pool = SynchronizedThreadPool()

        #: mp.Queue: Queue for retrieving events ('event_name', value) from model
        self.event_queue = mp.Queue(100)

        #: Manager: A shared memory manager
        self.manager = Manager()
        #: dict: Configuration dictionary
        self.configuration = load_configs(
            self.manager,
            configuration=configuration_path,
            experiment=experiment_path,
            waveform_constants=waveform_constants_path,
            rest_api_config=rest_api_path,
            waveform_templates=waveform_templates_path,
        )

        verify_experiment_config(self.manager, self.configuration)
        verify_waveform_constants(self.manager, self.configuration)

        # Initialize the Model
        #: ObjectInSubprocess: Model object in MVC architecture.
        self.model = ObjectInSubprocess(
            Model, args, self.configuration, event_queue=self.event_queue
        )

        logger.info(f"Spec - Configuration Path: {configuration_path}")
        logger.info(f"Spec - Experiment Path: {experiment_path}")
        logger.info(f"Spec - Waveform Constants Path: {waveform_constants_path}")
        logger.info(f"Spec - Rest API Path: {rest_api_path}")

        # Wire up pipes
        #: mp.Pipe: Pipe for sending images from model to view.
        self.show_img_pipe = self.model.create_pipe("show_img_pipe")

        # save default experiment file
        #: string: Path to the default experiment yaml file.
        self.default_experiment_file = experiment_path

        # waveform setting file
        #: string: Path to the waveform constants yaml file.
        self.waveform_constants_path = waveform_constants_path

        # Configuration Reader
        #: ConfigurationController: Configuration Controller object.
        self.configuration_controller = ConfigurationController(self.configuration)

        # Initialize the View
        #: View: View object in MVC architecture.
        self.view = view(root)

        # Sub Gui Controllers
        #: AcquireBarController: Acquire Bar Sub-Controller.
        self.acquire_bar_controller = AcquireBarController(
            self.view.acqbar, self
        )
        #: ChannelsTabController: Channels Tab Sub-Controller.
        self.channels_tab_controller = ChannelsTabController(
            self.view.settings.channels_tab, self
        )
        #: MultiPositionController: Multi-Position Tab Sub-Controller.
        self.multiposition_tab_controller = MultiPositionController(
            self.view.settings.multiposition_tab.multipoint_list, self
        )
        #: CameraViewController: Camera View Tab Sub-Controller.
        self.camera_view_controller = CameraViewController(
            self.view.camera_waveform.camera_tab, self
        )
        #: CameraSettingController: Camera Settings Tab Sub-Controller.
        self.camera_setting_controller = CameraSettingController(
            self.view.settings.camera_settings_tab, self
        )
        #: StageController: Stage Sub-Controller.
        self.stage_controller = StageController(
            self.view.settings.stage_control_tab,
            self.view,
            self.camera_view_controller.canvas,
            self,
        )
        #: WaveformTabController: Waveform Display Sub-Controller.
        self.waveform_tab_controller = WaveformTabController(
            self.view.camera_waveform.waveform_tab, self
        )
        #: KeystrokeController: Keystroke Sub-Controller.
        self.keystroke_controller = KeystrokeController(self.view, self)

        # Exit
        self.view.root.protocol(
            "WM_DELETE_WINDOW", self.acquire_bar_controller.exit_program
        )

        # Bonus config
        self.update_acquire_control()

        t = threading.Thread(target=self.update_event)
        t.start()

        # self.microscope = self.configuration['configuration']
        # ['microscopes'].keys()[0]  # Default to the first microscope

        # Initialize the menus
        #: MenuController: Menu Sub-Controller.
        self.menu_controller = MenuController(view=self.view, parent_controller=self)
        self.menu_controller.initialize_menus()
        #: dict: acquisition modes from plugins
        self.plugin_acquisition_modes = {}
        # add plugin menus
        #: PluginsController: Plugin Sub-Controller
        self.plugin_controller = PluginsController(view=self.view, parent_controller=self)
        self.plugin_controller.load_plugins()

        # Create default data buffer
        #: int: Number of x_pixels from microscope configuration file.
        self.img_width = 0
        #: int: Number of y_pixels from microscope configuration file.
        self.img_height = 0
        #: SharedNDArray: Pre-allocated shared memory array.
        self.data_buffer = None
        #: dict: Additional microscopes.
        self.additional_microscopes = {}
        #: dict: Additional microscope configurations.
        self.additional_microscopes_configs = {}
        #: bool: Flag for stopping acquisition.
        self.stop_acquisition_flag = False

        # Set view based on model.experiment
        self.populate_experiment_setting(in_initialize=True)

        # Camera View Tab
        self.initialize_cam_view()

        # destroy splash screen and show main screen
        splash_screen.destroy()
        root.deiconify()

        #: event: Event for resizing the GUI. Only works on Windows OS.
        self.resize_event_id = None
        if platform.system() == "Windows":
            self.view.root.bind("<Configure>", self.resize)

    def update_buffer(self):
        """Update the buffer size according to the camera
        dimensions listed in the experimental parameters.

        Returns
        -------
        self.img_width : int
            Number of x_pixels from microscope configuration file.
        self.image_height : int
            Number of y_pixels from microscope configuration file.
        self.data_buffer : SharedNDArray
            Pre-allocated shared memory array.
            Size dictated by x_pixels, y_pixels, an number_of_frames in
            configuration file.
        """
        img_width = int(
            self.configuration["experiment"]["CameraParameters"]["img_x_pixels"]
        )
        img_height = int(
            self.configuration["experiment"]["CameraParameters"]["img_y_pixels"]
        )
        if img_width == self.img_width and img_height == self.img_height:
            return

        self.data_buffer = self.model.get_data_buffer(img_width, img_height)
        self.img_width = img_width
        self.img_height = img_height

    def update_acquire_control(self):
        """Update the acquire control based on the current experiment parameters."""
        self.view.acqbar.stop_stage.config(
            command=self.stage_controller.stop_button_handler
        )

    def change_microscope(self, microscope_name, zoom=None):
        """Change the microscope configuration.

        Parameters
        ----------
        microscope_name : string
            Name of the microscope to change to.
        zoom : string
            Name of the zoom value to change to.
        """
        self.configuration["experiment"]["MicroscopeState"][
            "microscope_name"
        ] = microscope_name
        if zoom:
            self.configuration["experiment"]["MicroscopeState"]["zoom"] = zoom
        if self.configuration_controller.change_microscope():
            # update widgets
            self.stage_controller.initialize()
            self.channels_tab_controller.initialize()
            self.camera_setting_controller.update_camera_device_related_setting()
            self.camera_setting_controller.calculate_physical_dimensions()
            self.camera_view_controller.update_snr()

        if (
            hasattr(self, "waveform_popup_controller")
            and self.waveform_popup_controller
        ):
            self.waveform_popup_controller.populate_experiment_values()

    def initialize_cam_view(self):
        """Populate view tab.

        Populate widgets with necessary data from
        config file via config controller. For the entire view tab.
        Sets the minimum and maximum counts
        for when the data is not being autoscaled.
        """
        # Populating Min and Max Counts
        minmax_values = [0, 2**16 - 1]
        self.camera_view_controller.initialize("minmax", minmax_values)
        image_metrics = [1, 0, 0]
        self.camera_view_controller.initialize("image", image_metrics)

    def populate_experiment_setting(self, file_name=None, in_initialize=False):
        """Load experiment file and populate model.experiment and configure view.

        Confirms that the experiment file exists.
        Sends the experiment file to the model and the controller.
        Populates the GUI with these settings.

        Parameters
        __________
        file_name : string
            file_name = path to the non-default experiment yaml file.

        """
        # read the new file and update info of the configuration dict
        if not in_initialize:
            update_config_dict(
                self.manager, self.configuration, "experiment", file_name
            )
            verify_experiment_config(self.manager, self.configuration)

        # update buffer
        self.update_buffer()

        # Configure GUI
        microscope_name = self.configuration["experiment"]["MicroscopeState"][
            "microscope_name"
        ]
        self.configuration_controller.change_microscope()
        self.menu_controller.resolution_value.set(
            f"{microscope_name} "
            f"{self.configuration['experiment']['MicroscopeState']['zoom']}"
        )
        self.menu_controller.disable_stage_limits.set(
            0 if self.configuration["experiment"]["StageParameters"]["limits"] else 1
        )
        self.execute(
            "stage_limits",
            self.configuration["experiment"]["StageParameters"]["limits"],
        )

        self.acquire_bar_controller.populate_experiment_values()
        # self.stage_controller.populate_experiment_values()
        self.multiposition_tab_controller.set_positions(
            self.configuration["experiment"]["MultiPositions"]
        )
        self.channels_tab_controller.populate_experiment_values()
        self.camera_setting_controller.populate_experiment_values()

        # autofocus popup
        if hasattr(self, "af_popup_controller"):
            self.af_popup_controller.populate_experiment_values()

        if file_name:
            self.plugin_controller.populate_experiment_setting()

        # set widget modes
        self.set_mode_of_sub("stop")
        self.stage_controller.initialize()

    def update_experiment_setting(self):
        """Update model.experiment according to values in the GUI

        Collect settings from sub-controllers will validate the value, if something
        is wrong, it will return False

        Returns
        -------
        string
            Warning info if any

        """
        self.camera_setting_controller.update_experiment_values()
        # update multi-positions
        positions = self.multiposition_tab_controller.get_positions()
        update_config_dict(
            self.manager,
            self.configuration["experiment"],
            "MultiPositions",
            positions,
        )
        self.configuration["experiment"]["MicroscopeState"][
            "multiposition_count"
        ] = len(positions)

        # TODO: validate experiment dict

        channel_warning = self.channels_tab_controller.verify_experiment_values()
        if channel_warning:
            return channel_warning
        if self.configuration["experiment"]["MicroscopeState"]["scanrange"] == 0:
            return "Scan range shouldn't be 0!"
        return ""

    def resize(self, event):
        """Resize the GUI.

        Parameters
        __________
        event : event
            event = <Configure x=0 y=0 width=1200 height=600>
        """

        def refresh(width, height):
            """ Refresh the GUI.

            Parameters
            __________
            width : int
                Width of the GUI.
            height : int
                Height of the GUI.
            """
            if width < 1200 or height < 600:
                return
            self.view.camera_waveform["width"] = (
                width - self.view.frame_left.winfo_width() - 81
            )
            self.view.camera_waveform["height"] = height - 110

        if event.widget != self.view.scroll_frame:
            return
        if self.resize_event_id:
            self.view.after_cancel(self.resize_event_id)
        self.resize_event_id = self.view.after(
            1000, lambda: refresh(event.width, event.height)
        )

    def prepare_acquire_data(self):
        """Prepare the acquisition data.

        Updates model.experiment.
        Sets sub-controller's mode to 'live' when 'continuous is selected, or 'stop'.

        Returns
        -------
        bool
            True if all settings are valid, False otherwise.
        """
        warning_info = self.update_experiment_setting()
        if warning_info:
            messagebox.showerror(
                title="Warning",
                message=f"Cannot start acquisition!\n{warning_info}",
            )
            return False

        # update real image width and height
        self.set_mode_of_sub(self.acquire_bar_controller.mode)
        self.update_buffer()
        return True

    def set_mode_of_sub(self, mode):
        """Communicates imaging mode to sub-controllers.

        Parameters
        __________
        mode : string
            string = 'live', 'stop'
        """
        self.channels_tab_controller.set_mode(mode)
        self.camera_view_controller.set_mode(mode)
        self.camera_setting_controller.set_mode(mode)
        self.waveform_tab_controller.set_mode(mode)
        if mode == "stop":
            # GUI Failsafe
            self.acquire_bar_controller.stop_acquire()
            # self.menu_controller.feature_id_val.set(0)

    def execute(self, command, *args):
        """Functions listens to the Sub_Gui_Controllers.

        The controller.experiment is passed as an argument to the model, which then
        overwrites the model.experiment. Workaround due to model being in a sub-process.

        Parameters
        __________
        command : string
            string = 'stage', 'stop_stage', 'move_stage_and_update_info',
        args* : function-specific passes.
        """

        if command == "joystick_toggle":
            """Toggles the joystick mode on/off."""
            if self.stage_controller.joystick_is_on:
                self.execute("stop_stage")

        elif command == "stage":
            """Creates a thread and uses it to call the model to move stage

            Parameters
            __________
            args[0] : dict
                dict = {'x': value, 'y': value, 'z': value, 'theta': value, 'f': value}
            """
            self.threads_pool.createThread(
                "model", self.move_stage, args=({args[1] + "_abs": args[0]},)
            )

        elif command == "stop_stage":
            """Creates a thread and uses it to call the model to stop stage"""
            self.threads_pool.createThread("stop_stage", self.stop_stage)

        elif command == "move_stage_and_update_info":
            """update stage view to show the position

            Parameters
            __________
            args[0] : dict
                dict = {'x': value, 'y': value, 'z': value, 'theta': value, 'f': value}
            """
            self.stage_controller.set_position(args[0])

        elif command == "move_stage_and_acquire_image":
            """update stage and acquire an image

            Parameters
            __________
            args[0] : dict
                dict = {'x': value, 'y': value, 'z': value, 'theta': value, 'f': value}
            """
            stage_pos = dict(map(lambda axis: (axis + "_abs", args[0][axis]), args[0]))
            self.move_stage(stage_pos)
            self.update_stage_controller_silent(stage_pos)
            self.acquire_bar_controller.set_mode("single")
            self.execute("acquire")

        elif command == "get_stage_position":
            """Returns the current stage position

            Returns
            -------
                dict = {'x': value, 'y': value, 'z': value, 'theta': value, 'f': value}
            """
            return self.stage_controller.get_position()

        elif command == "mark_position":
            """Appends a position to the multi-position list.

            Parameters
            __________
            args[0] : dict
                dict = {'x': value, 'y': value, 'z': value, 'theta': value, 'f': value}
                values are in float64
            """
            self.multiposition_tab_controller.append_position(args[0])

        elif command == "resolution":
            """Changes the resolution mode and zoom position.

            Recalculates FOV_X and FOV_Y
            If Waveform Popup is open, communicates changes to it.

            Parameters
            ----------
            args : str
                "microscope_name zoom_value", "microscope_name", or "zoom_value"
            """
            # get microscope name and zoom value from args[0]
            temp = args[0].split()
            if len(temp) == 1:
                # microscope name is given
                if temp[0] in self.configuration_controller.microscope_list:
                    temp.append(
                        self.configuration_controller.get_zoom_value_list(temp[0])[0]
                    )
                elif temp[0] in self.configuration_controller.get_zoom_value_list(
                    self.configuration_controller.microscope_name
                ):
                    temp = [self.configuration_controller.microscope_name, temp[0]]
                else:
                    return
            resolution_value = " ".join(temp)
            if resolution_value != self.menu_controller.resolution_value.get():
                self.menu_controller.resolution_value.set(resolution_value)
                return
            self.change_microscope(temp[0], temp[1])
            work_thread = self.threads_pool.createThread(
                "model", lambda: self.model.run_command("update_setting", "resolution")
            )
            work_thread.join()

        elif command == "set_save":
            """Set whether the image will be saved.

            Parameters
            __________
            args : Boolean
                is_save = True/False
            """
            self.acquire_bar_controller.set_save_option(args[0])

        elif command == "update_setting":
            """Called by the Waveform Constants Popup Controller
            to update the Waveform constants settings in memory.

            Parameters
            __________
            args[0] : string
                string = 'resolution' or 'waveform' or 'galvo'...
            args[1] : dict
                dict = {
                'resolution_mode': self.resolution,
                'zoom': self.mag,
                'laser_info': self.resolution_info[
                'remote_focus_constants'][self.resolution][self.mag]
                }
            """
            self.threads_pool.createThread(
                "model", lambda: self.model.run_command("update_setting", *args)
            )

        elif command == "stage_limits":
            self.stage_controller.stage_limits = args[0]
            self.threads_pool.createThread(
                "model", lambda: self.model.run_command("stage_limits", *args)
            )

        elif command == "autofocus":
            """Execute autofocus routine."""
            if not self.acquire_bar_controller.is_acquiring:
                self.threads_pool.createThread(
                    "camera",
                    self.capture_image,
                    args=("autofocus", "live", *args),
                )
            elif self.acquire_bar_controller.mode == "live":
                self.threads_pool.createThread(
                    "model",
                    lambda: self.model.run_command("autofocus", *args)
                )

        elif command == "eliminate_tiles":
            """Execute eliminate tiles routine."""

            self.acquire_bar_controller.set_mode(mode="customized")
            feature_list = self.menu_controller.feature_list_names
            feature_name = "Remove Empty Tiles"
            try:
                # feature_id_val has a trace, and setting the menu item triggers it.
                feature_id = feature_list.index(feature_name)
                self.menu_controller.feature_id_val.set(feature_id)
            except ValueError:
                logger.debug("No feature named 'Remove Empty Tiles' found.")
                messagebox.showwarning(title="Navigate",
                                       message="Feature 'Remove Empty Tiles' not found."
                                       )
                return
            self.execute("acquire")

        elif command == "load_feature":
            """Tell model to load/unload features."""

            work_thread = self.threads_pool.createThread(
                "model", lambda: self.model.run_command("load_feature", *args)
            )
            work_thread.join()

        elif command == "acquire_and_save":
            """Acquire data and save it.

            Prepares the acquisition data.
            Creates the file directory for saving the data.
            Saves the experiment file to that directory.
            Acquires the data.

            Parameters
            __________
            args[0] : dict
                dict = self.save_settings from the experiment.yaml file.

            """
            if not self.prepare_acquire_data():
                self.acquire_bar_controller.stop_acquire()
                return
            saving_settings = self.configuration["experiment"]["Saving"]
            file_directory = create_save_path(saving_settings)
            save_yaml_file(
                file_directory,
                self.configuration["experiment"],
                filename="experiment.yml",
            )
            self.camera_setting_controller.solvent = self.configuration["experiment"][
                "Saving"
            ]["solvent"]
            self.camera_setting_controller.calculate_physical_dimensions()
            self.execute("acquire")

        elif command == "acquire":
            """Acquire data.

            Triggered when the Acquire button is hit by the user in the GUI.

            Prepares the acquisition data.

            Parameters
            __________
            args[0] : string
                string = 'continuous', 'z-stack', 'single', or 'projection'
            """
            # acquisition mode from plugin
            plugin_obj = self.plugin_acquisition_modes.get(
                self.acquire_bar_controller.mode, None)

            if plugin_obj and hasattr(plugin_obj, "prepare_acquisition_controller"):
                getattr(plugin_obj, "prepare_acquisition_controller")(self)

            # Prepare data
            if not self.prepare_acquire_data():
                self.acquire_bar_controller.stop_acquire()
                return
            
            # set the display segmentation flag to False
            self.camera_view_controller.display_mask_flag = False

            # ask user to verify feature list parameters if in "customized" mode
            if self.acquire_bar_controller.mode == "customized":
                feature_id = self.menu_controller.feature_id_val.get()
                if feature_id > 0:
                    if hasattr(self, "features_popup_controller"):
                        self.features_popup_controller.exit_func()
                    feature_list_popup = FeatureListPopup(
                        self.view, title="Feature List Configuration"
                    )
                    self.features_popup_controller = FeaturePopupController(
                        feature_list_popup, self
                    )
                    self.features_popup_controller.populate_feature_list(feature_id)
                    # wait until close the popup windows
                    self.view.wait_window(feature_list_popup.popup)
                    # do not run acquisition if "cancel" is selected
                    temp = self.features_popup_controller.start_acquisiton_flag
                    delattr(self, "features_popup_controller")
                    if not temp:
                        self.set_mode_of_sub("stop")
                        return

                    # if select 'ilastik segmentation' and 'show segmentation',
                    # TODO: update id if the feature id is changed
                    self.camera_view_controller.display_mask_flag = (
                        self.menu_controller.feature_id_val.get() == 4
                        and self.ilastik_controller.show_segmentation_flag
                    )

            self.stop_acquisition_flag = False
            self.launch_additional_microscopes()

            self.threads_pool.createThread(
                "camera",
                self.capture_image,
                args=(
                    "acquire",
                    self.acquire_bar_controller.mode,
                ),
            )

        elif command == "stop_acquire":
            """Stop the acquisition."""
            self.stop_acquisition_flag = True

            # self.model.run_command('stop')
            self.sloppy_stop()
            self.menu_controller.feature_id_val.set(0)

            # clear show_img_pipe
            while self.show_img_pipe.poll():
                # TODO: image_id never called.
                self.show_img_pipe.recv()
                # image_id = self.show_img_pipe.recv()

        elif command == "exit":
            """Exit the program."""
            # Save current GUI settings to .navigate/config/experiment.yml file.
            self.sloppy_stop()
            # self.menu_controller.feature_id_val.set(0)

            self.update_experiment_setting()
            file_directory = os.path.join(get_navigate_path(), "config")
            save_yaml_file(
                file_directory=file_directory,
                content_dict=self.configuration["experiment"],
                filename="experiment.yml",
            )
            if hasattr(self, "waveform_popup_controller"):
                self.waveform_popup_controller.save_waveform_constants()

            self.model.run_command("terminate")
            self.model = None
            self.event_queue.put(("stop", ""))
            self.threads_pool.clear()
            sys.exit()

        # mirror commands:
        elif command in [
            "flatten_mirror",
            "zero_mirror",
            "set_mirror",
            "set_mirror_from_wcs",
        ]:
            self.threads_pool.createThread(
                "model", lambda: self.model.run_command(command, *args)
            )
        elif command == "tony_wilson":
            self.threads_pool.createThread(
                "camera",
                self.capture_image,
                args=(
                    "tony_wilson",
                    "live",
                ),
            )
        else:
            self.threads_pool.createThread(
                "model", lambda: self.model.run_command(command, *args)
            )

        # elif command == "change_camera":
        #     self.model.run_command("change_camera", *args)

        logger.info(
            f"Navigate Controller - command passed from child, {command}, {args}"
        )

    def sloppy_stop(self):
        """Keep trying to stop the model until successful.

        TODO: Delete this function!!!

        This is set up to get around the conflict between
        self.threads_pool.createThread('model', target)
        commands and the need to stop as abruptly as
        possible when the user hits stop. Here we leverage
        ObjectInSubprocess's refusal to let us access
        the model from two threads to our advantage, and just
        try repeatedly until we get a command in front
        of the next command in the model threads_pool resource.
        We should instead pause the model thread pool
        and interject our stop command, or clear the queue
        in threads_pool.
        """
        e = RuntimeError
        while e == RuntimeError:
            try:
                self.model.run_command("stop")
                e = None
            except RuntimeError:
                e = RuntimeError

    def capture_image(self, command, mode, *args):
        """Trigger the model to capture images.

        Parameters
        ----------
        command : string
            string = 'acquire' or 'autofocus'
        mode : string
            string = 'continuous', 'z-stack', 'single', or 'projection'
        args : function-specific passes.
        """
        self.camera_view_controller.image_count = 0

        # Start up Progress Bars
        images_received = 0
        self.acquire_bar_controller.progress_bar(
            images_received=images_received,
            microscope_state=self.configuration["experiment"]["MicroscopeState"],
            mode=mode,
            stop=False,
        )
        try:
            work_thread = self.threads_pool.createThread(
                "model", lambda: self.model.run_command(command, *args)
            )
            work_thread.join()
        except Exception as e:
            messagebox.showerror(
                title="Error:",
                message=f"WARNING:\n{e}",
            )
            self.set_mode_of_sub("stop")
            return
        self.acquire_bar_controller.view.acquire_btn.configure(text="Stop")
        self.acquire_bar_controller.view.acquire_btn.configure(state="normal")

        self.camera_view_controller.initialize_non_live_display(
            self.data_buffer,
            self.configuration["experiment"]["MicroscopeState"],
            self.configuration["experiment"]["CameraParameters"],
        )

        self.stop_acquisition_flag = False

        while True:
            if self.stop_acquisition_flag:
                break
            # Receive the Image and log it.
            image_id = self.show_img_pipe.recv()
            logger.info(f"Navigate Controller - Received Image: {image_id}")

            if image_id == "stop":
                break
            if not isinstance(image_id, int):
                logger.debug(
                    f"Navigate Controller - Something wrong happened, stop the model!, "
                    f"{image_id}"
                )
                self.execute("stop_acquire")

            # Display the Image in the View
            self.camera_view_controller.try_to_display_image(image_id=image_id)
            images_received += 1

            # Update progress bar.
            self.acquire_bar_controller.progress_bar(
                images_received=images_received,
                microscope_state=self.configuration["experiment"]["MicroscopeState"],
                mode=mode,
                stop=False,
            )

        logger.info(
            f"Navigate Controller - Captured {images_received}, " f"{mode} Images"
        )

        # acquisition mode from plugin
        plugin_obj = self.plugin_acquisition_modes.get(mode, None)
        if plugin_obj and hasattr(plugin_obj, "end_acquisition_controller"):
            getattr(plugin_obj, "end_acquisition_controller")(self)

        # Stop Progress Bars
        self.acquire_bar_controller.progress_bar(
            images_received=images_received,
            microscope_state=self.configuration["experiment"]["MicroscopeState"],
            mode=mode,
            stop=True,
        )
        self.set_mode_of_sub("stop")

    def launch_additional_microscopes(self):
        """Launch additional microscopes."""

        def display_images(camera_view_controller, show_img_pipe, data_buffer):
            """Display images from additional microscopes.

            Parameters
            ----------
            camera_view_controller : CameraViewController
                Camera View Controller object.
            show_img_pipe : multiprocessing.Pipe
                Pipe for showing images.
            data_buffer : SharedNDArray
                Pre-allocated shared memory array.
                Size dictated by x_pixels, y_pixels, an number_of_frames in
                configuration file.
            """
            camera_view_controller.initialize_non_live_display(
                data_buffer,
                self.configuration["experiment"]["MicroscopeState"],
                self.configuration["experiment"]["CameraParameters"],
            )
            images_received = 0
            while True:
                if self.stop_acquisition_flag:
                    break
                # Receive the Image and log it.
                image_id = show_img_pipe.recv()
                logger.info(f"Navigate Controller - Received Image: {image_id}")

                if image_id == "stop":
                    break
                if not isinstance(image_id, int):
                    logger.debug(
                        f"Navigate Controller - Something wrong happened in additional "
                        f"microscope!, {image_id}"
                    )
                    break

                # Display the Image in the View
                try:
                    camera_view_controller.try_to_display_image(
                        image_id=image_id,
                    )
                except tkinter._tkinter.TclError:
                    print("Can't show images for the additional microscope!")
                    break
                images_received += 1

        # destroy unnecessary additional microscopes
        temp = []
        for microscope_name in self.additional_microscopes:
            if microscope_name not in self.additional_microscopes_configs:
                temp.append(microscope_name)
        for microscope_name in temp:
            del self.additional_microscopes[microscope_name]
            self.model.destroy_virtual_microscope(microscope_name)

        # show additional camera view popup
        for microscope_name in self.additional_microscopes_configs:
            if microscope_name not in self.additional_microscopes:
                show_img_pipe = self.model.create_pipe(
                    f"{microscope_name}_show_img_pipe"
                )
                data_buffer = self.model.launch_virtual_microscope(
                    microscope_name,
                    self.additional_microscopes_configs[microscope_name],
                )

                self.additional_microscopes[microscope_name] = {
                    "show_img_pipe": show_img_pipe,
                    "data_buffer": data_buffer,
                }
            if (
                self.additional_microscopes[microscope_name].get(
                    "camera_view_controller", None
                )
                is None
            ):
                popup_window = CameraViewPopupWindow(self.view, microscope_name)
                camera_view_controller = CameraViewController(
                    popup_window.camera_view, self
                )
                camera_view_controller.data_buffer = self.additional_microscopes[
                    microscope_name
                ]["data_buffer"]
                popup_window.popup.bind("<Configure>", camera_view_controller.resize)
                self.additional_microscopes[microscope_name][
                    "camera_view_controller"
                ] = camera_view_controller
                popup_window.popup.protocol(
                    "WM_DELETE_WINDOW",
                    combine_funcs(
                        popup_window.popup.dismiss,
                        lambda: self.additional_microscopes[microscope_name].pop(
                            "camera_view_controller"
                        ),
                    ),
                )

            # clear show_img_pipe
            show_img_pipe = self.additional_microscopes[microscope_name][
                "show_img_pipe"
            ]
            while show_img_pipe.poll():
                image_id = show_img_pipe.recv()
                if image_id == "stop":
                    break

            # start thread
            capture_img_thread = threading.Thread(
                target=display_images,
                args=(
                    self.additional_microscopes[microscope_name][
                        "camera_view_controller"
                    ],
                    show_img_pipe,
                    self.additional_microscopes[microscope_name]["data_buffer"],
                ),
            )
            capture_img_thread.start()

    def move_stage(self, pos_dict):
        """Trigger the model to move the stage.

        Parameters
        ----------
        pos_dict : dict
            Dictionary of axis positions
        """
        # Update our local stage dictionary
        update_stage_dict(self, pos_dict)

        # Pass to model
        self.model.move_stage(pos_dict)

    def stop_stage(self):
        """Stop the stage.

        Grab the stopped position from the stage
        and update the GUI control values accordingly.
        """
        self.model.stop_stage()

    def update_stage_controller_silent(self, ret_pos_dict):
        """Send updates to the stage GUI

        Parameters
        ----------
        ret_pos_dict : dict
            Dictionary of axis positions
        """
        stage_gui_dict = {}
        for axis, val in ret_pos_dict.items():
            ax = axis.split("_")[0]
            stage_gui_dict[ax] = val
        self.stage_controller.set_position_silent(stage_gui_dict)

    def update_event(self):
        """Update the View/Controller based on events from the Model."""
        while True:
            event, value = self.event_queue.get()

            if event == "warning":
                # Display a warning that arises from the model as a top-level GUI popup
                messagebox.showwarning(title="Navigate", message=value)

            elif event == "waveform":
                # Update the waveform plot.
                self.waveform_tab_controller.update_waveforms(
                    waveform_dict=value,
                    sample_rate=self.configuration_controller.daq_sample_rate,
                )
            elif event == "multiposition":
                # Update the multi-position tab without appending to the list
                update_table(
                    table=self.view.settings.multiposition_tab.multipoint_list.get_table(),
                    pos=value,
                )
                self.channels_tab_controller.is_multiposition_val.set(True)
                self.channels_tab_controller.toggle_multiposition()

            elif event == "disable_multiposition":
                self.channels_tab_controller.is_multiposition_val.set(False)
                self.channels_tab_controller.toggle_multiposition()

            elif event == "ilastik_mask":
                # Display the ilastik mask
                self.camera_view_controller.display_mask(mask=value)

            elif event == "autofocus":
                # Display the autofocus plot
                if hasattr(self, "af_popup_controller"):
                    self.af_popup_controller.display_plot(
                        data=value[0], line_plot=value[1], clear_data=value[2]
                    )

            elif event == "tonywilson":
                if hasattr(self, "ao_popup_controller"):
                    # self.ao_popup_controller.set_widgets_from_coef(value['coefs'])
                    self.ao_popup_controller.plot_tonywilson(value)
                    # self.ao_popup_controller.plot_mirror(value)
                    if value["done"]:
                        print("Tony Wilson done! Updating expt...")
                        self.ao_popup_controller.update_experiment_values()
            elif event == "mirror_update":
                if hasattr(self, "ao_popup_controller"):
                    self.ao_popup_controller.set_widgets_from_coef(value["coefs"])
                    self.ao_popup_controller.plot_mirror(value)
            elif event == "ao_save_report":
                if hasattr(self, "ao_popup_controller"):
                    self.ao_popup_controller.save_report_to_file(value)

            elif event == "stop":
                # Stop the software
                break

            elif event == "update_stage":
                # ZM: I am so sorry for this.
                for _ in range(10):
                    try:
                        self.update_stage_controller_silent(value)
                        break
                    except RuntimeError:
                        time.sleep(0.001)
                        pass

            elif event == "framerate":
                # Update the Framerate in the Camera Settings Tab
                self.camera_setting_controller.framerate_widgets["max_framerate"].set(
                    value
                )

                # Update the Framerate in the Acquire Bar to provide an estimate of
                # the duration of time remaining.
                self.acquire_bar_controller.framerate = value
            elif event == "remove_positions":
                self.multiposition_tab_controller.remove_positions(value)

    # def exit_program(self):
    #     """Exit the program.

    #     This function is called when the user clicks the exit button in the GUI.
    #     """
    #     if messagebox.askyesno("Exit", "Are you sure?"):
    #         logger.info("Exiting Program")
    #         self.execute("exit")
    #         sys.exit()

    def add_acquisition_mode(self, name, acquisition_obj):
        if name in self.plugin_acquisition_modes:
            print(f"*** plugin acquisition mode {name} exists, can't add another one!")
            return
        self.plugin_acquisition_modes[name] = acquisition_obj(name)
        self.acquire_bar_controller.add_mode(name)



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import tkinter as tk
from tkinter import ttk

# Local Imports
from navigate.view.custom_widgets.popup import PopUp
from navigate.view.custom_widgets.LabelInputWidgetFactory import LabelInput


class FeatureIcon(ttk.Button):
    """Feature Icon Widget"""

    def __init__(self, parent, feature_name="", set_bg=False):
        """Initialize the Feature Icon Widget

        Parameters
        ----------
        parent : tk.Frame
            Parent frame of the widget
        feature_name : str
            Name of the feature
        """
        ttk.Button.__init__(self, parent, text=feature_name)

        # Create a style
        style = ttk.Style()
        style.configure("Custom.TButton", background="red")
        if set_bg:
            self.configure(style="Custom.TButton")
        
        self.configure(padding=(10, 20))


class FeatureConfigPopup:
    """Feature Config Popup Widget"""

    def __init__(
        self,
        root,
        *args,
        features=[],
        feature_name="",
        args_name=[],
        args_value=[],
        **kwargs
    ):
        """Initialize the Feature Config Popup Widget

        Parameters
        ----------
        root : tk.Tk
            Root window of the application
        *args : list
            List of arguments
        features : list
            List of features
        feature_name : str
            Name of the feature
        args_name : list
            List of arguments name
        args_value : list
            List of arguments value
        """
        # Creating popup window with this name and size/placement,

        #: PopUp: Popup window
        self.popup = PopUp(
            root, kwargs["title"], "+320+180", top=False, transient=False
        )

        # Change background of popup window to white
        self.popup.configure(bg="white")

        # Creating content frame
        content_frame = self.popup.get_frame()

        #: list: List of input widgets
        self.inputs = []

        #: LabelInput: Feature name widget
        self.feature_name_widget = LabelInput(
            parent=content_frame,
            label="Feature Name",
            label_args={"width": 20},
            input_class=ttk.Combobox,
            input_var=tk.StringVar(),
            input_args={"width": 30, "state": "readonly"},
        )
        self.feature_name_widget.grid(row=0, column=0, sticky=tk.NSEW, padx=3, pady=3)
        self.feature_name_widget.set(feature_name)
        self.feature_name_widget.set_values(features)

        separator = ttk.Separator(content_frame)
        separator.grid(row=1, column=0, sticky=tk.NSEW, pady=10)

        #: ttk.Frame: Parameter frame
        self.parameter_frame = ttk.Frame(content_frame)
        self.parameter_frame.grid(row=2, column=0, sticky=tk.NSEW, padx=30, pady=30)

        row = 3
        if "true" in kwargs:
            self.preview_btn_true = ttk.Button(content_frame, text="Preview (True)")
            self.preview_btn_true.grid(row=row, column=0, sticky=tk.NSEW)
            separator = ttk.Separator(content_frame)
            separator.grid(row=row+1, column=0, sticky=tk.NSEW, pady=(0,10))
            self.feature_list_true_frame = FeatureListFrame(content_frame)
            self.feature_list_true_frame.grid(row=row+2, column=0, sticky=tk.NSEW)
            row += 3

        if "false" in kwargs:
            self.preview_btn_false = ttk.Button(content_frame, text="Preview (False)")
            self.preview_btn_false.grid(row=row, column=0, sticky=tk.NSEW)
            separator = ttk.Separator(content_frame)
            separator.grid(row=row+1, column=0, sticky=tk.NSEW, pady=(0,10))
            self.feature_list_false_frame = FeatureListFrame(content_frame)
            self.feature_list_false_frame.grid(row=row+2, column=0, sticky=tk.NSEW)

        self.build_widgets(args_name, args_value, kwargs.get("parameter_config", {}))

    def build_widgets(self, args_name, args_value, parameter_config=None):
        """Build widgets for the popup

        Parameters
        ----------
        args_name : list
            List of arguments name
        args_value : list
            List of arguments value
        parameter_config : dict
            Dictionary of parameter configuration
        """
        #: list: List of input widgets
        self.inputs = []
        #: list: List of input widgets type
        self.inputs_type = []

        for child in self.parameter_frame.winfo_children():
            child.destroy()

        for i, arg_name in enumerate(args_name):
            arg_input_class = ttk.Entry
            arg_input_var = tk.StringVar
            if type(args_value[i]) is bool:
                arg_input_class = ttk.Combobox
                values = ["True", "False"]
            elif parameter_config is not None and arg_name in parameter_config:
                arg_input_class = ttk.Combobox
                values = list(parameter_config[arg_name].keys())

            temp = LabelInput(
                parent=self.parameter_frame,
                label=arg_name + ":",
                label_args={"padding": (2, 5, 5, 0), "width": 20},
                input_class=arg_input_class,
                input_var=arg_input_var(),
                input_args={"width": 30},
            )

            self.inputs.append(temp)
            self.inputs_type.append(type(args_value[i]))
            temp.grid(row=i + 2, column=0, sticky=tk.NSEW, padx=30, pady=10)
            if arg_input_class is ttk.Combobox:
                temp.set_values(values)
                temp.widget.config(state="readonly")
            temp.set(str(args_value[i]))

    def get_widgets(self):
        """Get widgets

        Returns
        -------
        list
            List of input widgets
        """
        return self.inputs


class FeatureListPopup:
    """Feature List Popup Widget"""

    def __init__(self, root, *args, **kwargs):
        """Initialize the Feature List Popup Widget

        Parameters
        ----------
        root : tk.Tk
            Root window of the application
        *args : list
            List of arguments
        **kwargs : dict
            Dictionary of keyword arguments
        """
        # Creating popup window with this name and size/placement,
        # PopUp is a Toplevel window
        #: PopUp: Popup window
        self.popup = PopUp(
            root, kwargs["title"], "+320+180", top=False, transient=False
        )
        # Change background of popup window to white
        self.popup.configure(bg="white")
        #: bool: Flag to indicate if the popup is for adding new list
        self.add_new_list_flag = False
        if kwargs["title"].startswith("Add"):
            self.add_new_list_flag = True

        # Creating content frame
        content_frame = self.popup.get_frame()

        #: dict: Dictionary of input widgets
        self.inputs = {}
        self.inputs["feature_list_name"] = LabelInput(
            parent=content_frame,
            label="Feature List Name",
            input_class=ttk.Entry,
            input_var=tk.StringVar(),
            input_args={"width": 50},
        )

        self.inputs["feature_list_name"].grid(
            row=0, column=0, sticky=tk.NSEW, padx=3, pady=3
        )

        separator = ttk.Separator(content_frame)
        separator.grid(row=2, column=0, sticky=tk.NSEW, padx=3, pady=3)

        scroll_frame = ttk.Frame(content_frame)
        scroll_frame.grid(row=3, column=0, sticky=tk.NSEW)
        canvas = tk.Canvas(scroll_frame, width=800, height=350)
        scrollbar = ttk.Scrollbar(scroll_frame, orient="horizon", command=canvas.xview)
        self.feature_view_frame = ttk.Frame(canvas)

        self.feature_view_frame.bind(
            "<Configure>", lambda e: canvas.configure(scrollregion=canvas.bbox("all"))
        )

        canvas.create_window((0, 0), window=self.feature_view_frame, anchor="nw")

        canvas.configure(xscrollcommand=scrollbar.set)

        canvas.pack(side="top", fill="both", expand=True)
        scrollbar.pack(side="bottom", fill="x")

        separator = ttk.Separator(content_frame)
        separator.grid(row=4, column=0, sticky=tk.NSEW, padx=3, pady=3)
        self.inputs["content"] = tk.Text(content_frame, width=100, height=10)
        self.inputs["content"].grid(row=5, column=0, sticky=tk.NSEW, padx=10, pady=3)

        #: dict: Dictionary of buttons
        self.buttons = {}
        button_frame = ttk.Frame(content_frame)
        button_frame.grid(row=6, column=0, sticky=tk.NSEW)
        self.buttons["preview"] = ttk.Button(button_frame, text="Preview")
        self.buttons["preview"].grid(row=0, column=0, padx=3, pady=3)
        if self.add_new_list_flag:
            self.buttons["add"] = ttk.Button(button_frame, text="Add")
            self.buttons["add"].grid(row=0, column=1, padx=3, pady=3)
        else:
            self.buttons["confirm"] = ttk.Button(button_frame, text="Confirm")
            self.buttons["confirm"].grid(row=0, column=1, padx=3, pady=3)
        self.buttons["cancel"] = ttk.Button(button_frame, text="Cancel")
        self.buttons["cancel"].grid(row=0, column=2, sticky=tk.SE, padx=3, pady=3)

class FeatureListFrame(ttk.Frame):
    """Feature list graph frame"""
    def __init__(self, root, *args, width=800, height=200, **kwargs):
        super().__init__(root)

        scroll_frame = ttk.Frame(self)
        scroll_frame.grid(row=0, column=0, sticky=tk.NSEW)
        canvas = tk.Canvas(scroll_frame, width=width, height=height)
        scrollbar = ttk.Scrollbar(scroll_frame, orient="horizon", command=canvas.xview)
        self.feature_view_frame = ttk.Frame(canvas)

        self.feature_view_frame.bind(
            "<Configure>", lambda e: canvas.configure(scrollregion=canvas.bbox("all"))
        )

        canvas.create_window((0, 0), window=self.feature_view_frame, anchor="nw")

        canvas.configure(xscrollcommand=scrollbar.set)

        canvas.pack(side="top", fill="both", expand=True)
        scrollbar.pack(side="bottom", fill="x")

        self.content = tk.Text(self, width=100, height=5)
        self.content.grid(row=2, column=0, sticky=tk.NSEW, padx=10, pady=3)



class FeatureAdvancedSettingPopup:
    """Feature Advanced Setting Popup Widget"""

    def __init__(
        self,
        root,
        *args,
        features=[],
        feature_name="",
        args_name=[],
        args_default_value=[],
        **kwargs
    ):
        """Initialize the Feature Advanced Setting Popup Widget

        Parameters
        ----------
        root : tk.Tk
            Root window of the application
        features : list
            List of features
        feature_name : str
            Name of the feature
        args_name : list
            List of arguments name
        args_default_value : list
            List of arguments default value
        """
        # Creating popup window with this name and size/placement,
        # PopUp is a Toplevel window
        #: PopUp: Popup window
        self.popup = PopUp(
            root, kwargs["title"], "+320+180", top=False, transient=False
        )
        # Change background of popup window to white
        self.popup.configure(bg="white")

        # Creating content frame
        content_frame = self.popup.get_frame()

        #: dict: Dictionary of input widgets
        self.inputs = {}
        #: dict: Dictionary of buttons
        self.buttons = {}

        self.feature_name_widget = LabelInput(
            parent=content_frame,
            label="Feature Name",
            label_args={"width": 20},
            input_class=ttk.Combobox,
            input_var=tk.StringVar(),
            input_args={"width": 30, "state": "readonly"},
        )
        self.feature_name_widget.grid(row=0, column=0, sticky=tk.NSEW, padx=3, pady=3)
        self.feature_name_widget.set(feature_name)
        self.feature_name_widget.set_values(features)

        separator = ttk.Separator(content_frame)
        separator.grid(row=1, column=0, sticky=tk.NSEW, pady=10)

        #: ttk.Frame: Parameter frame
        self.parameter_frame = ttk.Frame(content_frame)
        self.parameter_frame.grid(row=2, column=0, sticky=tk.NSEW, padx=30, pady=30)
        #: dict: Dictionary of argument frames
        self.arg_frames = {}

    def build_widgets(self, args_name, parameter_config=None):
        """Build widgets for the popup

        Parameters
        ----------
        args_name : list
            List of arguments name
        parameter_config : dict
            Dictionary of parameter configuration
        """
        #: dict: Dictionary of input widgets
        self.inputs = {}
        #: dict: Dictionary of buttons
        self.buttons = {}
        #: dict: Dictionary of argument frames
        self.arg_frames = {}

        for child in self.parameter_frame.winfo_children():
            child.destroy()
        self.save_button = None

        row_id = 0
        for _, arg_name in enumerate(args_name):
            # ref_value, value, delete button, load button
            self.inputs[arg_name] = []
            arg_label = ttk.Label(self.parameter_frame, text=arg_name + ":")
            arg_label.grid(row=row_id, column=0, sticky=tk.NW)
            row_id += 1
            arg_frame = ttk.Frame(self.parameter_frame)
            arg_frame.grid(row=row_id, column=0, sticky=tk.NSEW)
            row_id += 1
            self.arg_frames[arg_name] = arg_frame
            if parameter_config is not None and arg_name in parameter_config:
                for k, v in parameter_config[arg_name].items():
                    self.add_new_row(arg_name, k, v)
            add_button = ttk.Button(self.parameter_frame, text="Add")
            add_button.grid(row=row_id, column=0, sticky=tk.NW, padx=3, pady=3)
            self.buttons[arg_name] = add_button
            row_id += 1
            separator = ttk.Separator(self.parameter_frame)
            separator.grid(row=row_id, column=0, sticky=tk.NSEW, pady=10)
            row_id += 1

        if len(args_name) > 1:
            save_button = ttk.Button(self.parameter_frame, text="Save")
            save_button.grid(row=row_id, column=0, sticky=tk.NE, padx=3, pady=3)
            #: ttk.Button: Save button
            self.save_button = save_button

    def add_new_row(self, arg_name, k="", v=""):
        """Add new row to the popup

        Parameters
        ----------
        arg_name : str
            Name of the argument
        k : str
            Name of the function
        v : str
            Value of the function
        """
        r = len(self.inputs[arg_name])
        arg_frame = self.arg_frames[arg_name]
        ref_value_entry = LabelInput(
            parent=arg_frame,
            label="Function Name",
            input_class=ttk.Entry,
            input_var=tk.StringVar(),
            input_args={"width": 50},
        )
        ref_value_entry.grid(row=r, column=1, sticky=tk.NSEW, padx=3, pady=3)
        ref_value_entry.set(k)
        value_entry = LabelInput(
            parent=arg_frame,
            label="Value",
            input_class=ttk.Entry,
            input_var=tk.StringVar(),
            input_args={"width": 50},
        )
        value_entry.grid(row=r, column=2, sticky=tk.NSEW, padx=3, pady=3)
        value_entry.set(v)
        load_button = ttk.Button(arg_frame, text="Load")
        load_button.grid(row=r, column=3, sticky=tk.NSEW, padx=3, pady=3)
        delete_button = ttk.Button(arg_frame, text="Delete")
        delete_button.grid(row=r, column=4, sticky=tk.NSEW, padx=3, pady=3)
        delete_button.config(command=self.delete_row(arg_name, r))
        self.inputs[arg_name].append(
            (ref_value_entry, value_entry, load_button, delete_button)
        )

    def delete_row(self, arg_name, r):
        """Delete row from the popup

        Parameters
        ----------
        arg_name : str
            Name of the argument
        r : int
            Row number
        """

        def func():
            """Function to delete row from the popup

            Returns
            -------
            func
                Function to delete row from the popup
            """
            for w in self.inputs[arg_name][r]:
                w.grid_remove()
            self.inputs[arg_name][r] = None

        return func



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import logging

# Third Party Imports
import numpy as np
from navigate.model.features.image_writer import ImageWriter

# from navigate.model import data_sources
# from navigate.model.data_sources import data_source
# # from navigate.model.data_sources import D
# import os


p = __name__.split(".")[1]
logger = logging.getLogger(p)

# Local imports


class ConstantVelocityAcquisition:
    def __init__(self, model, axis="z", saving_flag=False, saving_dir="cva"):
        self.model = model

        self.axis = axis
        self.default_speed = None
        self.asi_stage = None
        self.saving_flag = saving_flag

        self.image_writer = None
        if self.saving_flag:
            self.image_writer = ImageWriter(model=self.model, sub_dir=saving_dir)

        self.config_table = {
            "signal": {
                "init": self.pre_func_signal,
                "main": self.main_func_signal,
                "end": self.end_func_signal,
                "cleanup": self.cleanup,
            },
            "data": {
                "init": self.pre_data_func,
                "main": self.in_data_func,
                "end": self.end_data_func,
                "cleanup": self.cleanup_data_function,
            },
            "node": {"node_type": "multi-step", "device_related": True},
        }

    def pre_func_signal(self):
        """Prepare the constant velocity acquisition.

        Injects a new trigger source and prepares the stage for a constant
        scan velocity mode. All tasks are triggered by the ASI stage
        encoder. Only operates in a per-stack acquisition mode.  Stage
        velocity is calculated based on the current exposure time,
        step size, and minimum encoder divide.

        Rotary encoder has a resolution of 45,396 counts/mm, or 1 count per
        10 nm. It is a quadrature device, so the minimum step size is 40 nm.

        Assumes stage motion is 45 degrees relative to the optical axis.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """

        print("**** CVA TTL STARTED pre signal func initiated****")
        self.asi_stage = self.model.active_microscope.stages[self.axis]
        microscope_state = self.model.configuration["experiment"]["MicroscopeState"]
        self.stack_cycling_mode = microscope_state["stack_cycling_mode"]
        self.channels = microscope_state["selected_channels"]
        self.current_channel_in_list = 0
        print(f"current channel = {self.channels}")

        self.end_acquisition = False
        self.received_frames = 0
        self.end_signal_temp = 0

        # get the current exposure time for channel channel.
        readout_time = self.model.active_microscope.get_readout_time()
        print("readout time calculated")
        print(f"*** readout time = {readout_time} s")
        (
            exposure_times,
            sweep_times,
        ) = self.model.active_microscope.calculate_exposure_sweep_times(readout_time)
        print("sweep times and exposure times calculated")
        print(f"exposure time = {exposure_times}")
        print(f"sweep time = {sweep_times}")
        channel_name = next(iter(sweep_times))

        channel_num = int(channel_name.split("_")[1])
        self.model.active_microscope.current_channel = channel_num
        print(f"channel_{self.model.active_microscope.current_channel}")
        current_sweep_time = sweep_times[
            f"channel_{self.model.active_microscope.current_channel}"
        ]
        current_expsure_time = exposure_times[
            f"channel_{self.model.active_microscope.current_channel}"
        ]
        self.current_sweep_time = current_sweep_time
        self.current_exposure_time = current_expsure_time
        self.readout_time = readout_time
        print("sweep time calculated")
        scaling_factor = 1
        print("*** current sweep time before scaling:", current_sweep_time)

        # Provide just a bit of breathing room for the sweep time...
        current_sweep_time = current_sweep_time * scaling_factor

        print("*** current sweep time after scaling:", current_sweep_time)
        print(f"*** Scaling Factor = {scaling_factor}")
        logger.info(f"*** current sweep time: {current_sweep_time}")
        logger.info(f"*** sweep time scaling: {scaling_factor}")

        # Calculate Stage Velocity
        encoder_resolution = 10  # nm
        minimum_encoder_divide = encoder_resolution * 4  # nm

        # Get step size from the GUI. For now, assume 160 nm.
        # Default units should probably be microns I believe. Confirm.
        # desired_sampling = 160  # nm
        desired_sampling = (
            float(
                self.model.configuration["experiment"]["MicroscopeState"]["step_size"]
            )
            * 1000.0
        )

        desired_sampling_um = float(
            self.model.configuration["experiment"]["MicroscopeState"]["step_size"]
        )

        print("*** step size um:", desired_sampling_um)
        logger.info(f"*** step size um: {desired_sampling_um}")

        # The stage is at 45 degrees relative to the optical axes.
        step_size = (desired_sampling * 2) / np.sqrt(2)  # 45 degrees, 226 nm
        print("Desired Axial Sampling:", desired_sampling)
        print("Desired Step Size of Stage:", step_size)

        # Calculate the desired encoder divide. Must be multiple of
        # minimum_encoder_divide. 2.6 encoder divides, round up to 3.
        # *** WHY IS THIS DIVIDE BY 2? TO CORRECT STEP SIZE ABOVE?
        desired_encoder_divide = np.ceil(step_size / minimum_encoder_divide)
        print("*** desired encoder divide:", desired_encoder_divide)
        # Calculate the actual step size in nanometers. 264 nm.
        step_size_nm = desired_encoder_divide * minimum_encoder_divide
        print("Actual Step Size of Stage:", step_size_nm)

        # Calculate the actual step size in millimeters. 264 * 10^-6 mm
        step_size_mm = step_size_nm * 10**-6  # 264 * 10^-6 mm

        # Set the start and end position of the scan in millimeters.
        # Retrieved from the GUI.
        # Set Stage Limits - Units in millimeters
        # microns to mm
        self.start_position = (
            float(
                self.model.configuration["experiment"]["MicroscopeState"]["abs_z_start"]
            )
            / 1000.0
        )
        self.stop_position = (
            float(
                self.model.configuration["experiment"]["MicroscopeState"]["abs_z_end"]
            )
            / 1000.0
        )
        self.number_z_steps = float(
            self.model.configuration["experiment"]["MicroscopeState"]["number_z_steps"]
        )

        self.start_position_um = self.start_position * 1000
        self.stop_position_um = self.stop_position * 1000

        logger.info(f"*** z start position: {self.start_position}")
        logger.info(f"*** z end position: {self.stop_position}")
        logger.info(f"*** Expected number of steps: {self.number_z_steps}")
        pos = self.asi_stage.get_axis_position(self.axis)
        print(f"Current Position = {pos}")
        print(f"Start position = {self.start_position*1000}")

        self.step_size_mm = step_size_mm
        self.asi_stage.scanr(
            start_position_mm=self.start_position,
            end_position_mm=self.stop_position,
            enc_divide=self.step_size_mm,
            axis=self.axis,
        )
        print("scan r initalized update channel")

        self.asi_stage.wait_until_complete(self.axis)
        print("Stage wait until complete completed after scanr")

        # TODO set max speed in configuration file
        # max_speed = 4.288497*2
        self.asi_stage.set_speed(percent=1)
        max_speed = self.asi_stage.get_speed(axis=self.axis)
        print(f"Axis {self.axis} Maximum Speed (mm/s): {max_speed}")
        logger.debug(f"Axis {self.axis} Maximum Speed (mm/s): {max_speed}")

        self.asi_stage.set_speed(percent=0.0001 / max_speed)
        basic_speed = self.asi_stage.get_speed(self.axis)  # mm/s
        print("Basic Speed = ", basic_speed)

        expected_speed = step_size_mm / current_sweep_time
        print("Expected velocity = ", expected_speed)

        self.expected_speed = expected_speed
        self.max_speed = max_speed
        self.percent_speed = self.expected_speed / self.max_speed

        self.asi_stage.set_speed(percent=expected_speed / max_speed)
        stage_velocity = self.asi_stage.get_speed(self.axis)
        print("Final stage velocity, (mm/s):", stage_velocity)
        print("Encoder divide step size = ", step_size_mm)
        logger.info(f"*** Expected stage velocity, (mm/s): {expected_speed}")
        logger.info(f"*** Final stage velocity, (mm/s): {stage_velocity}")

        actual_mechanical_step_size_mm = stage_velocity * self.current_sweep_time
        self.actual_mechanical_step_size_um = actual_mechanical_step_size_mm * 1000
        new_z_steps = np.floor(
            abs(self.stop_position_um - self.start_position_um)
            / self.actual_mechanical_step_size_um
        )

        expected_frames_v1 = np.ceil(
            ((new_z_steps * step_size_mm) / stage_velocity) / current_sweep_time
        )
        expected_frames = int(
            np.ceil(
                abs(self.start_position - self.stop_position)
                / stage_velocity
                / current_sweep_time
            )
        )
        print(f"*** Expected Frames from zsteps:{expected_frames_v1}")
        print(f"*** Expected Frames from start and stop position: {expected_frames}")
        print(f"*** new_z_steps = {new_z_steps}")
        logger.info(f"*** Expected Frames: {expected_frames}")
        self.model.configuration["experiment"]["MicroscopeState"][
            "waveform_template"
        ] = "CVACONPRO"
        self.model.configuration["waveform_templates"]["CVACONPRO"][
            "expand"
        ] = expected_frames
        self.model.configuration["waveform_templates"]["CVACONPRO"]["repeat"] = 1

        self.repeat_waveform = self.model.configuration["waveform_templates"][
            "CVACONPRO"
        ]["repeat"]
        self.expand_waveform = self.model.configuration["waveform_templates"][
            "CVACONPRO"
        ]["expand"]
        print(f"repeat num = {self.repeat_waveform}")
        print(f"expand num = {self.expand_waveform}")
        expand_frames = float(
            self.model.configuration["waveform_templates"]["CVACONPRO"]["expand"]
        )
        self.expected_frames = expected_frames
        # np.ceil(expected_frames/(self.repeat_waveform*self.expand_waveform))
        print(f"Expand Frames = {expand_frames}")
        print(f"Self Expected Frames test = {self.expected_frames}")
        logger.info(f"Self Expected Frames test = {self.expected_frames}")
        logger.info(f"Expand Frames = {expand_frames}")
        self.model.configuration["experiment"]["MicroscopeState"][
            "number_z_steps"
        ] = expected_frames

        # Updates metadata for saving each channel with the correct number of frames
        # self.file_type = self.model.configuration["experiment"]["Saving"]["file_type"]
        # self.data_source = data_sources.get_data_source(self.file_type)
        # self.data_source.set_metadata_from_configuration_experiment(self.model.configuration)

        self.model.active_microscope.current_channel = 0
        self.waveform_dict = self.model.active_microscope.calculate_all_waveform()
        self.model.active_microscope.daq.external_trigger = "/PXI6259/PFI1"
        print(
            f"external trigger set to "
            f"{self.model.active_microscope.daq.external_trigger}"
        )
        self.model.active_microscope.prepare_next_channel()

        posw = self.asi_stage.get_axis_position(self.axis)
        print(f"current position = {posw}, start position = {self.start_position_um}")

        self.model.resume_data_thread()
        print("DATA THREAD RESUMED")

    def main_func_signal(self):
        print("main function started")
        self.asi_stage.set_speed(percent=self.percent_speed)
        print("original stage speed set")
        posw = self.asi_stage.get_axis_position(self.axis)
        print(f"current position = {posw}, start position = {self.start_position_um}")
        self.asi_stage.start_scan(self.axis)
        print("scan started")

    def end_func_signal(self):
        print("in end_func_signal")

        if self.model.stop_acquisition:
            print("PRESSED STOP BUTTON")
            return True

        print(f"self.recieved frames = {self.received_frames}")
        print(f"model.stop_acquisition = {self.model.stop_acquisition}")
        print(f"self.end_acquisition = {self.end_acquisition}")
        pos = self.asi_stage.get_axis_position(self.axis)
        print(f"pos = {pos} stop position:{self.stop_position*1000}")
        self.end_signal_temp += 1
        print(f"self.end_signal_temp = {self.end_signal_temp}")
        print(f"DAQ Trigger SENT test = {self.end_signal_temp>0}")
        if (
            self.model.stop_acquisition
            or self.end_acquisition
            or self.end_signal_temp > 0
        ):
            print("returning true from stop or end acquisition end_func_signal")
            if self.stack_cycling_mode == "per_stack":
                print("per stack if statment called")
                print(f"channel list: {self.current_channel_in_list}")
                self.update_channel()
                print("if statement update channel finished")
                # if run through all the channels, move to next position
                if self.current_channel_in_list == 0:
                    print(
                        f"in if channel list = 0 statement, channel = "
                        f"{self.current_channel_in_list}"
                    )
                    return True
            else:
                print("in else true statement")
                return True

        print("returning False from end_func_signal")
        return False

    def update_channel(self):
        print("update channel")
        print(f"channel before in update channel = {self.channels}")
        print(f"channel before in update channel list = {self.current_channel_in_list}")
        self.current_channel_in_list = (
            self.current_channel_in_list + 1
        ) % self.channels
        print(f"channel after in update channel = {self.channels}")
        print(f"channel before in update channel list = {self.current_channel_in_list}")
        self.asi_stage.wait_until_complete(self.axis)
        print("Stage wait until complete completed")

        if self.current_channel_in_list > 0:
            print("PAUSING DATA THREAD IF STATEMENT IN CHANNEL")
            self.model.pause_data_thread()
            print("DATA THREAD PAUSED IN CHANNEL")

        self.asi_stage.set_speed(percent=0.5)
        print("Stage speed set")
        self.asi_stage.wait_until_complete(self.axis)
        print("Stage wait until complete completed")

        if self.current_channel_in_list == 0:
            print("if current channel is zero statement")
            return True

        self.asi_stage.set_speed(percent=self.percent_speed)
        print("original stage speed set")

        print(f"current channel = {self.model.active_microscope.current_channel}")
        self.model.active_microscope.prepare_next_channel()
        print("channel prepared after model")

        self.model.resume_data_thread()
        print("DATA THREAD RESUMED IN CHANNEL")

    def cleanup(self):
        """Clean up the constant velocity acquisition.

        Need to reset the trigger source to the default.

        """
        # reset stage speed
        # 4.288497*2
        pos = self.asi_stage.get_axis_position(self.axis)
        print(f"Current Position = {pos}")
        print(f"Stop position = {self.stop_position*1000}")
        print("Clean up called")
        # self.asi_stage.set_speed({self.axis: self.default_speed})
        self.asi_stage.set_speed(percent=0.5)

        # self.asi_stage.set_speed(0.9)
        print("speed set")
        end_speed = self.asi_stage.get_speed(self.axis)  # mm/s
        print("end Speed = ", end_speed)
        self.asi_stage.stop()
        print("stage stop")
        # print("external trigger none")
        # return to start position
        # self.start_position = float(
        #     self.model.configuration[
        #         "experiment"]["MicroscopeState"]["abs_z_start"])
        # self.asi_stage.move_absolute({f"{self.axis}_abs: {self.start_position}"})

        # self.asi_stage.move_axis_absolute(self.axis,
        # self.start_position * 1000.0, wait_until_done=True)
        # print("stage moved to original position")
        # pos = self.asi_stage.get_axis_position(self.axis)
        # print(f"Current Position = {pos}")
        # print(f"start position = {self.start_position*1000}")

        self.asi_stage.wait_until_complete(self.axis)
        print("Stage wait until complete completed after update scan")

        # self.asi_stage.stop_scan()
        # print("stage stop scan")
        print("DAQ STOP ACQUISITION CALLED CLEAN UP")
        self.model.active_microscope.daq.stop_acquisition()
        print("DAQ STOP ACQUISITION FINISHED CLEAN UP")
        self.model.configuration["experiment"]["MicroscopeState"][
            "waveform_template"
        ] = "Default"
        print("config set to default")
        self.model.active_microscope.current_channel = 0
        print(
            f"self.model.active_microscope.current_channel = "
            f"{self.model.active_microscope.current_channel})"
        )
        self.model.active_microscope.daq.external_trigger = None
        print("external trigger set to none")
        print("DAQ set external trigger to none called")
        # self.model.active_microscope.daq.set_external_trigger(None)
        print("external trigger set to none v2")
        # self.model.active_microscope.prepare_next_channel()
        # print("clean up channel prepared")
        # self.model.active_microscope.daq.set_external_trigger()
        # print("external trigger set to none v2")

    def pre_data_func(self):
        self.received_frames_v2 = self.received_frames
        self.total_frames = self.expected_frames * self.channels
        print(f"total channels = {self.channels}")
        print(f"total frames = {self.total_frames}")

    def in_data_func(self, frame_ids):
        # print(f"frame_ids = {len(frame_ids)}")
        self.received_frames += len(frame_ids)
        if self.image_writer is not None:
            self.image_writer.save_image(frame_ids)
        # self.received_frames_v2 = self.received_frames
        # self.recieved_frames_v2 += self.recieved_frames
        # print(f"received_Frames v2: {self.recieved_frames_v2}")

    def end_data_func(self):
        pos = self.asi_stage.get_axis_position(self.axis)
        expected_channel = self.expected_frames * (self.end_signal_temp + 1)
        # time_total = (self.total_frames * self.current_sweep_time) / 60
        # time_per_channel = (expected_channel * self.current_sweep_time) / 60
        time_remaining_total = round(
            ((self.total_frames - self.received_frames) * self.current_sweep_time)
            / (60),
            2,
        )
        time_remaining_per_channel = round(
            ((expected_channel - self.received_frames) * self.current_sweep_time)
            / (60),
            2,
        )
        print(
            f"Received: {self.received_frames} Per Channel: {expected_channel} "
            f"Expected Total: {self.total_frames} Channel = {self.end_signal_temp + 1}"
            f" of {self.channels}"
        )
        print(
            f"time remaining per channel min: {time_remaining_per_channel}, time "
            f"remaining total min: {time_remaining_total}"
        )
        logger.info(
            f"Received: {self.received_frames} Per Channel: {expected_channel} "
            f"Expected Total: {self.total_frames}"
        )
        # print(f"Received V2: {self.received_frames_v2} " \
        #        "Expected: {self.expected_frames}")
        print(f"Position: {pos} Stop Position: {self.stop_position*1000} ")
        logger.info(f"Position: {pos} Stop Position: {self.stop_position*1000} ")
        self.end_acquisition = self.received_frames >= self.total_frames
        return self.end_acquisition

    def cleanup_data_function(self):
        """Clean up the constant velocity acquisition.

        Designed to be called in the case of a failure.
        Do not rely on this function being called.

        Cleans up the image writer.
        """
        print("clean up data function called")
        if self.image_writer:
            self.image_writer.cleanup()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import logging

# Third Party Imports
import nidaqmx
from nidaqmx.constants import LineGrouping

# Local Imports
from navigate.model.devices.shutter.laser_shutter_base import ShutterBase

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class ShutterTTL(ShutterBase):
    """ShutterTTL Class

    Triggering for shutters delivered from DAQ using digital outputs.
    Each output keeps their last digital state for as long the device is not
    powered down.
    """

    def __init__(self, microscope_name, device_connection, configuration):
        """Initialize the ShutterTTL.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : object
            Hardware device to connect to
        configuration : multiprocessing.managers.DictProxy
            Global configuration of the microscope
        """
        super().__init__(microscope_name, device_connection, configuration)

        shutter_channel = configuration["configuration"]["microscopes"][
            microscope_name
        ]["shutter"]["hardware"]["channel"]

        #: nidaqmx.Task: NI Data Acquisition Digital Output Task
        self.shutter_task = nidaqmx.Task()
        self.shutter_task.do_channels.add_do_chan(
            shutter_channel, line_grouping=LineGrouping.CHAN_FOR_ALL_LINES
        )
        self.shutter_task.write(self.shutter_state, auto_start=True)

    def __del__(self):
        """Close the ShutterTTL at exit."""
        self.shutter_task.close()

    def open_shutter(self):
        """Open the shutter"""
        #: bool: Shutter state
        self.shutter_state = True
        self.shutter_task.write(self.shutter_state, auto_start=True)
        logger.debug("ShutterTTL - Shutter opened")

    def close_shutter(self):
        """Close the shutter"""
        self.shutter_state = False
        self.shutter_task.write(self.shutter_state, auto_start=True)
        logger.debug("ShutterTTL - The shutter is closed")

    @property
    def state(self):
        """Return the state of both shutters

        Returns
        -------
        shutter_state : bool
            State of the shutter.
        """
        return self.shutter_state



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Standard Library Imports
import logging

# Third Party Imports

# Local Imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class ConfigurationController:
    """Configuration Controller - Used to get the configuration of the microscope."""

    def __init__(self, configuration):
        """Initialize the Configuration Controller

        Parameters
        ----------
        configuration : dict
            The configuration dictionary.
        """
        #: dict: The configuration dictionary.
        self.configuration = configuration

        #: str: The microscope name.
        self.microscope_name = None

        #: dict: The microscope configuration dictionary.
        self.microscope_config = None
        self.change_microscope()

        microscopes_config = configuration["configuration"]["microscopes"]

        #: int: The number of galvos.
        self.galvo_num = max(
            map(
                lambda microscope_name: len(
                    microscopes_config[microscope_name]["galvo"]
                ),
                microscopes_config.keys(),
            )
        )

    def change_microscope(self) -> bool:
        """Get the new microscope configuration dict according to the name.

        Gets the name of the microscope, retrieves its configuration, and updates the
        Configuration Controller's attributes.

        Returns
        -------
        result: bool
        """
        microscope_name = self.configuration["experiment"]["MicroscopeState"][
            "microscope_name"
        ]
        assert (
            microscope_name in self.configuration["configuration"]["microscopes"].keys()
        )

        if self.microscope_name == microscope_name:
            return False

        self.microscope_config = self.configuration["configuration"]["microscopes"][
            microscope_name
        ]
        self.microscope_name = microscope_name
        return True

    def get_microscope_configuration_dict(self):
        """Return microscope configuration dictionary

        Returns
        -------
        microscope_configuration_dict: dict
        """
        return self.microscope_config

    @property
    def channels_info(self):
        """Return the channels info

        Populate the channel combobox with the channels
        that are available in the configuration

        Returns
        -------
        setting : dict
            Channel settings, e.g. {
                'laser': ['488nm', '561nm', '642nm'],
                'filter': ['Empty-Alignment', 'GFP - FF01-515/30-32', '...}
        """
        if self.microscope_config is None:
            return {}

        setting = {
            "laser": self.lasers_info,
            "filter": list(
                self.microscope_config["filter_wheel"]["available_filters"].keys()
            ),
        }
        return setting

    @property
    def lasers_info(self):
        """Return the lasers info

        Populate the laser combobox with the lasers
        that are available in the configuration

        Returns
        -------
        laser_list : list
            List of lasers, e.g. ['488nm', '561nm', '642nm']
        """
        if self.microscope_config is None:
            return []

        return [
            str(laser["wavelength"]) + "nm"
            for laser in self.microscope_config["lasers"]
        ]

    @property
    def camera_config_dict(self):
        """Get camera configuration dict

        Returns
        -------
        camera_setting: dict
            Camera Setting, e.g. {

            }
        """
        if self.microscope_config is not None:
            return self.microscope_config["camera"]
        return None

    @property
    def camera_pixels(self):
        """Get default pixel values from camera

        Returns
        -------
        x_pixels : int
            Number of x pixels
        y_pixels : int
            Number of y pixels
        """
        if self.microscope_config is None:
            return [2048, 2048]

        return [
            self.microscope_config["camera"]["x_pixels"],
            self.microscope_config["camera"]["y_pixels"],
        ]

    @property
    def stage_default_position(self):
        """Get current position of the stage

        Returns
        -------
        position : dict
            Dictionary with x, y, z, theta, and f positions.
        """
        if self.microscope_config is not None:
            stage_position = self.microscope_config["stage"]["position"]
            position = {
                "x": stage_position["x_pos"],
                "y": stage_position["y_pos"],
                "z": stage_position["z_pos"],
                "theta": stage_position["theta_pos"],
                "f": stage_position["f_pos"],
            }
        else:
            position = {"x": 0, "y": 0, "z": 0, "theta": 0, "f": 0}
        return position

    @property
    def stage_step(self):
        """Get the step size of the stage

        Returns
        -------
        steps : dict
            Step size in x (same step size for y), z, theta, and f.
        """
        if self.microscope_config is not None:
            stage_dict = self.microscope_config["stage"]
            steps = {
                "x": stage_dict["x_step"],
                "y": stage_dict["y_step"],
                "z": stage_dict["z_step"],
                "theta": stage_dict["theta_step"],
                "f": stage_dict["f_step"],
            }
        else:
            steps = {"x": 10, "y": 10, "z": 10, "theta": 10, "f": 10}
        return steps

    def get_stage_position_limits(self, suffix):
        """Return the position limits of the stage

        Parameters
        ----------
        suffix : str
            '_min' or '_max'

        Returns
        -------
        position_limits : dict
            Depending on suffix, min or max stage limits, e.g.
            {'x': 2000, 'y': 2000, 'z': 2000, 'theta': 0, 'f': 2000}.

        """
        axis = ["x", "y", "z", "theta", "f"]
        if self.microscope_config is not None:
            stage_dict = self.microscope_config["stage"]
            position_limits = {}
            for a in axis:
                position_limits[a] = stage_dict[a + suffix]
        else:
            for a in axis:
                position_limits[a] = 0 if suffix == "_min" else 100
        return position_limits

    @property
    def stage_flip_flags(self):
        """Return the flip flags of the stage

        Returns
        -------
        flip_flags : dict
            {'x': bool, 'y': bool, 'z': bool, 'theta': bool, 'f': bool}.

        """
        if self.microscope_config is not None:
            stage_dict = self.microscope_config["stage"]
        else:
            stage_dict = {}
        flip_flags = {}
        for axis in ["x", "y", "z", "theta", "f"]:
            flip_flags[axis] = stage_dict.get(f"flip_{axis}", False)
        return flip_flags

    @property
    def camera_flip_flags(self):
        """Return the flip flags of the camera

        Returns
        -------
        flip_flags : dict
            {'x': bool, 'y': bool}.
        """
        if self.microscope_config is not None:
            camera_dict = self.microscope_config["camera"]
        else:
            camera_dict = {}
        flip_flags = {
            "x": camera_dict.get("flip_x", False),
            "y": camera_dict.get("flip_y", False),
        }
        return flip_flags

    @property
    def remote_focus_dict(self):
        """Return delay_percent, pulse_percent.

        Returns
        -------
        remote_focus_parameters : dict
            Dictionary with the remote focus percent delay and pulse percent.
        """
        if self.microscope_config is not None:
            return self.microscope_config["remote_focus_device"]

        return None

    @property
    def galvo_parameter_dict(self):
        """Return galvo parameter dict.

        Returns
        -------
        galvo_parameter_dict : dict
            Dictionary with the galvo parameters.
        """
        if self.microscope_config is not None:
            # Inject names into unnammed galvos
            for i, galvo in enumerate(self.microscope_config["galvo"]):
                if galvo.get("name") is None:
                    self.microscope_config["galvo"][i]["name"] = f"Galvo {i}"
            return self.microscope_config["galvo"]
        return None

    @property
    def daq_sample_rate(self):
        """Return daq sample rate.

        Returns
        -------
        daq_sample_rate : float
            Sample rate of the daq.
        """
        if self.microscope_config is not None:
            return self.microscope_config["daq"]["sample_rate"]
        return 100000

    @property
    def filter_wheel_setting_dict(self):
        """Return filter wheel setting dict.

        Returns
        -------
        filter_wheel_setting_dict : dict
            Dictionary with the filter wheel settings.
        """
        if self.microscope_config is not None:
            return self.microscope_config["filter_wheel"]
        return None

    @property
    def stage_setting_dict(self):
        """Return stage setting dict.

        Returns
        -------
        stage_setting_dict : dict
            Dictionary with the stage settings.
        """
        if self.microscope_config is not None:
            return self.microscope_config["stage"]
        return None

    @property
    def number_of_channels(self):
        """Return number of channels.

        Returns
        -------
        number_of_channels : int
            Number of channels.
        """
        if self.microscope_config is not None:
            return self.configuration[
                "configuration"]["gui"]["channels"].get("count", 5)
        return 5

    @property
    def microscope_list(self):
        """Return a list of microscope names

        Returns
        -------
        microscope_list : list
            List of microscope names.
        """
        return list(self.configuration["configuration"]["microscopes"].keys())

    def get_zoom_value_list(self, microscope_name):
        """Return a list of zoom values

        Returns
        -------
        zoom_value_list : list
            List of zoom values.
        """
        return self.configuration["waveform_constants"]["remote_focus_constants"][
            microscope_name
        ].keys()
    
    @property
    def gui_setting(self):
        return self.configuration["configuration"]["gui"]



# Multiprocessing to spread CPU load, threading for concurrency:
import multiprocessing as mp
import threading

# Printing from a child process is tricky:
import io
from contextlib import redirect_stdout

# Handling exceptions from a child process/thread is tricky:
import sys
import traceback
import inspect

# Making sure objects are cleaned up nicely is tricky:
import weakref

# Making sure a child process closes when the parent exits is tricky:
import atexit
import signal

import logging
import numpy as np

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)

# Sharing memory between child processes is tricky:
try:
    from multiprocessing import shared_memory
except ImportError:
    print("Warning - multiprocessing.shared_memory module could not be imported.")
    print("Shared memory arrays not implemented.")
    shared_memory = np.ndarray([0])
    # np = None

""" Created by Nathaniel H. Thayer and Andrew G. York

Full docstring at https://github.com/AndrewGYork/tools/blob/master/concurrency_tools.py.
"""


class SharedNDArray(np.ndarray):
    """A numpy array that lives in shared memory

    Inputs and outputs to/from ObjectInSubprocess are 'serialized', which
    is pretty fast - except for large in-memory objects. The only large
    in-memory objects we regularly deal with are numpy arrays, so it
    makes sense to provide a way to pass large numpy arrays via shared memory
    (which avoids slow serialization).

    Maybe you wanted to write code that looks like this:

        data_buf = np.zeros((400, 2000, 2000), dtype='uint16')
        display_buf = np.zeros((2000, 2000), dtype='uint8')

        camera = Camera()
        preprocessor = Preprocessor()
        display = Display()

        camera.record(num_images=400, out=data_buf)
        preprocessor.process(in=data_buf, out=display_buf)
        display.show(display_buf)

    ...but instead you write code that looks like this:

        data_buf = SharedNDArray(shape=(400, 2000, 2000), dtype='uint16')
        display_buf = SharedNDArray(shape=(2000, 2000), dtype='uint8')

        camera = ObjectInSubprocess(Camera)
        preprocessor = ObjectInSubprocess(Preprocessor)
        display = ObjectInSubprocess(Display)

        camera.record(num_images=400, out=data_buf)
        preprocessor.process(in=data_buf, out=display_buf)
        display.show(display_buf)

    ...and your payoff is, each object gets its own CPU core, AND passing
    large numpy arrays between the processes is still really fast!

    To implement this we used memmap from numpy.core as a template.
    """

    def __new__(
        cls,
        shape=None,
        dtype=float,
        shared_memory_name=None,
        offset=0,
        strides=None,
        order=None,
    ):
        if shared_memory_name is None:
            dtype = np.dtype(dtype)
            requested_bytes = np.prod(shape, dtype="uint64") * dtype.itemsize
            requested_bytes = int(requested_bytes)
            try:
                shm = shared_memory.SharedMemory(create=True, size=requested_bytes)
            except OSError as e:
                if e.args == (24, "Too many open files"):
                    raise OSError(
                        "You tried to simultaneously open more "
                        "SharedNDArrays than are allowed by your system!"
                    ) from e
                else:
                    raise e
            must_unlink = True  # This process is responsible for unlinking
        else:
            shm = shared_memory.SharedMemory(name=shared_memory_name, create=False)
            must_unlink = False
        obj = super(SharedNDArray, cls).__new__(
            cls, shape, dtype, shm.buf, offset, strides, order
        )
        obj.shared_memory = shm
        obj.offset = offset
        if must_unlink:
            weakref.finalize(obj, shm.unlink)
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        if not isinstance(obj, SharedNDArray):
            raise ValueError("You can't view non-shared memory as shared memory.")
        if hasattr(obj, "shared_memory") and np.may_share_memory(self, obj):
            self.shared_memory = obj.shared_memory
            self.offset = obj.offset
            self.offset += (
                self.__array_interface__["data"][0] - obj.__array_interface__["data"][0]
            )

    def __array_wrap__(self, arr, context=None):
        arr = super().__array_wrap__(arr, context)

        # Return a SharedNDArray if a SharedNDArray was given as the
        # output of the ufunc. Leave the arr class unchanged if self is not
        # a SharedNDArray to keep original SharedNDArray subclasses
        # behavior.

        if self is arr or type(self) is not SharedNDArray:
            return arr
        # Return scalar instead of 0d SharedMemory, e.g. for np.sum with
        # axis=None
        if arr.shape == ():
            return arr[()]
        # Return ndarray otherwise
        return arr.view(np.ndarray)

    def __getitem__(self, index):
        res = super().__getitem__(index)
        if type(res) is SharedNDArray and not hasattr(res, "shared_memory"):
            return res.view(type=np.ndarray)
        return res

    def __reduce__(self):
        args = (
            self.shape,
            self.dtype,
            self.shared_memory.name,
            self.offset,
            self.strides,
            None,
        )
        return (SharedNDArray, args)


class ResultThread(threading.Thread):
    """threading.Thread with all the simple features we wish it had.

    We added a 'get_result' method that returns values/raises exceptions.

    We changed the return value of 'start' from 'None' to 'self' -- just to
    trivially save us a line of code when launching threads.

    Example:
    ```
        def f(a):
            ''' A function that does something... '''
            return a.sum()

        ##
        ## Getting Results:
        ##
        a = np.ones((2,), dtype='uint8')

        # Our problem:
        th = threading.Thread(target=f, args=(a,))
        th.start()
        th.join() # We can't access the result of f(a) without redefining f!

        # Our solution:
        res_th = ResultThread(target=f, args=(a,)).start()
        res = res_th.get_result() # returns f(a)
        assert res == 2

        ##
        ## Error handling
        ##
        a = 1

        # Our problem:
        th = threading.Thread(target=f, args=(a,))
        th.start()
        th.join()
        # f(a) raised an unhandled exception. Our parent thread has no idea!

        # Our solution:
        res_th = ResultThread(target=f, args=(a,)).start()
        try:
            res = res_th.get_result()
        except AttributeError:
            print("AttributeError was raised in thread!")
        else:
            raise AssertionError(
                'We expected an AttributeError to be raised on join!')

        # Unhandled exceptions raised during evaluation of 'f' are reraised in
        # the parent thread when you call 'get_result'.
        # Tracebacks may print to STDERR when the exception occurs in
        # the child thread, but don't affect the parent thread (yet).
    ```
    NOTE: This module modifies threading.excepthook. You can't just copy/paste
    this class definition and expect it to work.
    """

    def __init__(self, group=None, target=None, name=None, args=(), kwargs=None):
        super().__init__(group, target, name, args, kwargs)
        self._return = None

    def run(self):
        if self._target is not None:
            self._return = self._target(*self._args, **self._kwargs)

    def start(self):
        try:
            super().start()
        except RuntimeError as e:
            if e.args == ("can't start new thread",):
                print("*" * 80)
                print("Failed to launch a thread.")
                print(threading.active_count(), "threads are currently active.")
                print("You might have reached a limit of your system;")
                print("let some of your threads finish before launching more.")
                print("*" * 80)
            raise
        return self

    def get_result(self, timeout=None):
        """Either returns a value or raises an exception.

        Optionally accepts a timeout in seconds. If thread has not returned
        after timeout seconds, raises a TimeoutError.
        """
        super().join(timeout=timeout)
        if self.is_alive():  # Thread could potentially not be done yet!
            return TimeoutError("Thread did not return!")
        if hasattr(self, "exc_value"):
            raise self.exc_value
        return self._return


class CustodyThread(ResultThread):
    """Threads that can access shared resources in the order they were launched.

    See the docstring at the top of this module for examples.
    """

    def __init__(
        self,
        first_resource=None,
        group=None,
        target=None,
        name=None,
        args=(),
        kwargs=None,
    ):
        if "custody" not in inspect.signature(target).parameters:
            raise ValueError(
                "The function 'target' passed to a CustodyThread"
                " must accept an argument named 'custody'"
            )
        custody = _Custody()  # Useful for synchronization in the launched thread
        if first_resource is not None:
            # Get in line for custody of the first resource the launched
            # thread will use, but don't *wait* in that line; the launched
            # thread should do the waiting, not the main thread:
            custody.switch_from(None, first_resource, wait=False)
        if kwargs is None:
            kwargs = {}
        if "custody" in kwargs:
            raise ValueError(
                "CustodyThread will create and pass a keyword argument to"
                " 'target' named 'custody', so keyword arguments to a"
                " CustodyThread can't be named 'custody'"
            )
        kwargs["custody"] = custody
        super().__init__(group, target, name, args, kwargs)
        self.custody = custody


_original_threading_excepthook = threading.excepthook


def _my_threading_excepthook(args):
    """Show a traceback when a child exception isn't handled by the parent."""
    if isinstance(args.thread, ResultThread):
        args.thread.exc_value = args.exc_value
        args.thread.exc_traceback = args.exc_traceback
        args.thread.exc_type = args.exc_type
    else:
        _try_to_print_child_traceback(args.exc_value)
    return _original_threading_excepthook(args)


threading.excepthook = _my_threading_excepthook

FancyThread = ResultThread  # So Andy can refer to it like this.
PoliteThread = CustodyThread


class ObjectInSubprocess:
    def __init__(
        self,
        initializer,
        *initargs,
        custom_loop=None,
        close_method_name=None,
        closeargs=None,
        closekwargs=None,
        with_lock=False,
        **initkwargs,
    ):
        """
        Make an object in a child process, that acts like it isn't.

        As much as possible, we try to make instances of ObjectInSubprocess
        behave as if they're an instance of the object living in the parent
        process. They're not, of course: they live in a child process. If you
        have spare cores on your machine, this turns CPU-bound operations
        (which  threading can't parallelize) into IO-bound operations (which
        threading CAN parallelize),  without too much mental overhead for the
        coder.

        initializer -- callable that returns an instance of a Python object
        initargs, initkwargs --  arguments to 'initializer'
        close_method_name -- string, optional, name of our object's method to
            be called automatically when the child process exits
        closeargs, closekwargs -- arguments to 'close_method'
        """
        # Put an instance of the Python object returned by 'initializer'
        # in a child process:
        parent_pipe, child_pipe = mp.Pipe()
        child_loop = _child_loop if custom_loop is None else custom_loop
        child_process = mp.Process(
            target=child_loop,
            name=initializer.__name__,
            args=(
                child_pipe,
                initializer,
                initargs,
                initkwargs,
                close_method_name,
                closeargs,
                closekwargs,
            ),
        )
        # Attribute-setting looks weird here because we override __setattr__,
        # and because we use a dummy object's namespace to hold our attributes
        # so we shadow as little of the object's namespace as possible:
        super().__setattr__("_", _DummyClass())  # Weird, but for a reason.
        self._.parent_pipe = parent_pipe
        self._.parent_pipe_lock = _ObjectInSubprocessPipeLock()
        self._.child_pipe = child_pipe
        self._.child_process = child_process
        self._.waiting_list = _WaitingList()
        if with_lock:
            self._.resource_lock = threading.Lock()
        else:
            self._.resource_lock = None
        # Make sure the child process initialized successfully:
        with self._.parent_pipe_lock:
            self._.child_process.start()
            assert _get_response(self) == "Successfully initialized"
        # Try to ensure the child process closes when we exit:
        dummy_namespace = getattr(self, "_")
        weakref.finalize(self, _close, dummy_namespace)
        try:
            signal.signal(signal.SIGTERM, lambda s, f: _close(dummy_namespace))
        except ValueError:  # We are probably starting from a thread.
            pass  # Signal handling can only happen from main thread

    def __getattr__(self, name):
        """Access attributes of the child-process object in the parent process.

        As much as possible, we want attribute access and method calls
        to *seem* like they're happening in the parent process, if
        possible, even though they actually involve asking the child
        process over a pipe.
        """
        if self._.resource_lock:
            self._.resource_lock.acquire()
        if name == "terminate":
            with self._.parent_pipe_lock:
                self._.parent_pipe.send(None)
                self._.child_process.terminate()
            if self._.resource_lock:
                self._.resource_lock.release()
            return _dummy_function
        with self._.parent_pipe_lock:
            self._.parent_pipe.send(("__getattribute__", (name,), {}))
            attr = _get_response(self)
        if callable(attr):

            def attr(*args, **kwargs):
                with self._.parent_pipe_lock:
                    self._.parent_pipe.send((name, args, kwargs))
                    return _get_response(self, True)

        elif self._.resource_lock:
            self._.resource_lock.release()
        return attr

    def __setattr__(self, name, value):
        with self._.parent_pipe_lock:
            self._.parent_pipe.send(("__setattr__", (name, value), {}))
            return _get_response(self)


def _get_response(object_in_subprocess, release=False):
    """
    Effectively a method of ObjectInSubprocess, but defined externally to
    minimize shadowing of the object's namespace
    """
    resp, printed_output = object_in_subprocess._.parent_pipe.recv()
    if len(printed_output) > 0:
        print(printed_output, end="")
    if isinstance(resp, Exception):
        raise resp
    if (
        release
        and object_in_subprocess._.resource_lock
        and object_in_subprocess._.resource_lock.locked()
    ):
        object_in_subprocess._.resource_lock.release()
    return resp


def _close(dummy_namespace):
    """Externally defined close function.

    Effectively a method of ObjectInSubprocess, but defined externally to
    minimize shadowing of the object's namespace
    """
    if not dummy_namespace.child_process.is_alive():
        return
    with dummy_namespace.parent_pipe_lock:
        dummy_namespace.parent_pipe.send(None)
        dummy_namespace.child_process.join()
        dummy_namespace.parent_pipe.close()


def _child_loop(
    child_pipe,
    initializer,
    initargs,
    initkwargs,
    close_method_name,
    closeargs,
    closekwargs,
):
    """The event loop of a ObjectInSubprocess's child process

    Parameters
    ----------
    child_pipe : object
        multiprocessing.connection.Connection object.
    initializer :
        ...
    initargs : tuple
        e.g., (False, Namespace(verbose=False, synthetic_hardware=True, debug=False,
        CPU=True, config_file=None, experiment_file=None, etl_const_file=None,
        logging_config=None))
    initkwargs : dict
        e.g., {'configuration_path': PosixPath('.../config/configuration.yml'),
        'experiment_path':
        PosixPath('.../config/experiment.yml'),
        'waveform_constants_path': PosixPath('.../config/waveform_constants.yml'),
        'event_queue': <multiprocessing.queues.Queue object at 0x7fdd509c7c40>}
    close_method_name : NoneType
        None
    closeargs : NoneType
        None
    closekwargs : NoneType
        None

    """
    # Initialization.
    printed_output = io.StringIO()
    try:  # Create an instance of our object...
        with redirect_stdout(printed_output):
            obj = initializer(*initargs, **initkwargs)
            if close_method_name is not None:
                close_method = getattr(obj, close_method_name)
                closeargs = tuple() if closeargs is None else closeargs
                closekwargs = dict() if closekwargs is None else closekwargs
                atexit.register(lambda: close_method(*closeargs, **closekwargs))
                # Note: We don't know if print statements in the close method
                # will print in the main process.
        child_pipe.send(("Successfully initialized", printed_output.getvalue()))
    except Exception as e:  # If we fail to initialize, just give up.
        e.child_traceback_string = traceback.format_exc()
        child_pipe.send((e, printed_output.getvalue()))
        return None
    # Main loop:
    while True:
        printed_output = io.StringIO()
        try:
            cmd = child_pipe.recv()
        except EOFError:  # This implies the parent is dead; exit.
            return None
        if cmd is None:  # This is how the parent signals us to exit.
            return None
        method_name, args, kwargs = cmd
        try:
            with redirect_stdout(printed_output):
                result = getattr(obj, method_name)(*args, **kwargs)
            if callable(result):
                result = _dummy_function  # Cheaper than sending a real callable
            child_pipe.send((result, printed_output.getvalue()))
        except Exception as e:
            # e.child_traceback_string = traceback.format_exc()
            print("Exception inside ObjectInSubprocess:", traceback.format_exc())
            child_pipe.send((Exception(str(e)), printed_output.getvalue()))


# A minimal class that we use just to get another namespace:


class _DummyClass:
    pass


# If we're trying to return a (presumably worthless) "callable" to
# the parent, it might as well be small and simple:


def _dummy_function():
    return None


class _WaitingList:
    """For synchronization of one-thread-at-a-time shared resources

    Each ObjectInSubprocess has a _WaitingList; if you want to define your own
    _WaitingList-like objects that can interact with
    _Custody.switch_from() and _Custody._wait_in_line(), make sure they have
    a waiting_list = [] attribute, and a waiting_list_lock =
    threading.Lock() attribute.
    """

    def __init__(self):
        self.waiting_list = []  # Switch to a queue/deque if speed really matters
        self.waiting_list_lock = threading.Lock()

    def __enter__(self):
        self.waiting_list_lock.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.waiting_list_lock.release()


class _ObjectInSubprocessPipeLock:
    """Raises an educational exception (rather than blocking) when you try
    to acquire a locked lock.
    """

    def __init__(self):
        self.lock = threading.Lock()

    def __enter__(self):
        if not self.lock.acquire(blocking=False):
            raise RuntimeError(
                "Two different threads tried to use the same "
                "ObjectInSubprocess at the same time! This is bad. Look at the "
                "docstring of concurrency_tools.py to see an example of how "
                "to use a _Custody object to avoid this problem."
            )
        return self.lock

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.lock.release()


threading_lock_type = type(threading.Lock())  # Used for typechecking


def _get_list_and_lock(resource):
    """Convenience function.

    Expected input: An ObjectInSubprocess, a _WaitingList, or a
    _WaitingList-like object with 'waiting_list' and
    'waiting_list_lock' attributes.
    """
    if isinstance(resource, ObjectInSubprocess):
        waiting_list = resource._.waiting_list.waiting_list
        waiting_list_lock = resource._.waiting_list.waiting_list_lock
    else:  # Either a _WaitingList, or a good enough impression
        waiting_list = resource.waiting_list
        waiting_list_lock = resource.waiting_list_lock
    assert isinstance(waiting_list_lock, threading_lock_type)
    return waiting_list, waiting_list_lock


class _Custody:
    def __init__(self):
        """For synchronization of single-thread-at-a-time shared resources.

        See the docstring at the start of this module for example usage.
        For _Custody() to be useful, at least some of the objects
        accessed by your launched thread must be ObjectInSubprocess()s,
        _WaitingList()s, or _WaitingList-like objects.
        """
        self.permission_slip = threading.Lock()
        self.permission_slip.acquire()
        self.has_custody = False
        self.target_resource = None

    def switch_from(self, resource, to=None, wait=True):
        """Get in line for a shared resource, then abandon your current resource

        If wait==True, also wait in that line until it's your turn to
        own the next shared resource.
        """
        assert resource is not None or to is not None
        if to is not None:
            to_waiting_list, to_waiting_list_lock = _get_list_and_lock(to)
            with to_waiting_list_lock:  # Get in the line for the next lock...
                if self not in to_waiting_list:  # ...unless you're already in it
                    to_waiting_list.append(self)
        if resource is not None:
            assert self.has_custody
            waiting_list, waiting_list_lock = _get_list_and_lock(resource)
            with waiting_list_lock:
                waiting_list.pop(0)  # Remove ourselves from the current line
                if len(waiting_list) > 0:  # If anyone's next...
                    # ...wake them up
                    waiting_list[0].permission_slip.release()
        self.has_custody = False
        self.target_resource = to
        if wait and self.target_resource is not None:
            self._wait_in_line()

    def release(self):
        """Release custody of the current shared resource.

        If you get custody of a shared resource and then raise an exception,
        the next-in-line might wait forever.

        'release' is useful while handling exceptions, if you want to pass
        custody of the resource to the next-in-line.

        This only works if you currently have custody, but it's hard to raise
        an exception while waiting in line.
        """
        if self.has_custody:
            self.switch_from(self.target_resource, to=None)
        else:
            if self.target_resource is None:
                return
            waiting_list, waiting_list_lock = _get_list_and_lock(self.target_resource)
            with waiting_list_lock:
                waiting_list.remove(self)

    def _wait_in_line(self):
        """Wait in line until it's your turn."""
        waiting_list, _ = _get_list_and_lock(self.target_resource)
        if self.has_custody:
            assert self is waiting_list[0]
            return
        # Wait for your number to be called
        if self is waiting_list[0] and self.permission_slip.locked():
            self.permission_slip.release()  # We arrived to an empty waiting list
        self.permission_slip.acquire()  # Blocks if we're not first in line
        self.has_custody = True


# When an exception from a child process isn't handled by the parent
# process, we'd like the parent to print the child traceback. Overriding
# sys.excepthook and threading.excepthook seems to be the standard way
# to do this:


def _try_to_print_child_traceback(v):
    if hasattr(v, "child_traceback_string"):
        print(
            f'{" Child Process Traceback ":v^79s}\n',
            v.child_traceback_string,
            f'{" Child Process Traceback ":^^79s}\n',
            f'{" Main Process Traceback ":v^79s}',
        )


def _my_excepthook(t, v, tb):
    """Show a traceback when a child exception isn't handled by the parent."""
    _try_to_print_child_traceback(v)
    return sys.__excepthook__(t, v, tb)


sys.excepthook = _my_excepthook

# Multiprocessing code works fairly differently depending whether you
# use 'spawn' or 'fork'. Since 'spawn' seems to be available on every
# platform we care about, and 'fork' is either missing or broken on some
# platforms, we'll always use 'spawn'. If your code calls
# mp.set_start_method() and sets it to anything other than 'spawn', this
# will crash with a RuntimeError. If you really need 'fork', or
# 'forkserver', then you probably know what you're doing better than us,
# and you shouldn't be using this module.
if mp.get_start_method(allow_none=True) != "spawn":
    mp.set_start_method("spawn")



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Standard Library Imports
import logging
from tkinter import ttk
import tkinter as tk

# Third Party Imports

# Local Imports
from navigate.view.custom_widgets.popup import PopUp

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class PluginsPopup:
    """Class creates the popup to uninstall plugins."""

    def __init__(self, root, *args, **kwargs):
        """Initialize the PluginsPopup class

        Parameters
        ----------
        root : tk.Tk
            The root window
        *args
            Variable length argument list
        **kwargs
            Arbitrary keyword arguments
        """
        # Creating popup window with this name and size/placement, PopUp is a Toplevel
        # window
        #: PopUp: The popup window
        self.popup = PopUp(
            root, "Plugins", "+320+180", transient=True
        )

        # Storing the content frame of the popup, this will be the parent of the widgets
        content_frame = self.popup.get_frame()

        # Formatting
        tk.Grid.columnconfigure(content_frame, "all", weight=1)
        tk.Grid.rowconfigure(content_frame, "all", weight=1)

        # Dictionary for all the variables
        #: dict: Variable
        self.variables = {}

        label = ttk.Label(content_frame, text="Plugin Name", width=30)
        label.grid(row=0, column=1, padx=(30, 10), sticky=tk.NE)

        label = ttk.Label(content_frame, text="Location", width=60)
        label.grid(row=0, column=2, sticky=tk.NE)

        separator = ttk.Separator(content_frame)
        separator.grid(row=1, columnspan=3, sticky=tk.NSEW)

        self.plugins_frame = ttk.Frame(content_frame)
        self.plugins_frame.grid(row=2, columnspan=3, sticky=tk.NSEW)
        
        self.uninstall_btn = ttk.Button(content_frame, text="Uninstall")
        self.uninstall_btn.grid(row=3, column=2, sticky=tk.NE)


    def build_widgets(self, plugin_config):
        """List all plugins"""
        self.variables = []
        for child in self.plugins_frame.winfo_children():
            child.destroy()

        for i, plugin_name in enumerate(plugin_config.keys()):
            var = tk.StringVar()
            check = ttk.Checkbutton(self.plugins_frame, variable=var, onvalue=plugin_name, offvalue="")
            check.grid(row=i, column=0, sticky=tk.NW)
            label = ttk.Label(self.plugins_frame, text=plugin_name)
            label.grid(row=i, column=1, padx=(5,10), sticky=tk.NSEW)
            label = ttk.Label(self.plugins_frame, text=plugin_config[plugin_name])
            label.grid(row=i, column=2, sticky=tk.NSEW)
            self.variables.append(var)



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library imports
import importlib
import logging
import time
from multiprocessing.managers import ListProxy

# Local Imports
from navigate.model.devices.stages.stage_base import StageBase


# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


def build_TLKIMStage_connection(serialnum):
    """Connect to the Thorlabs KIM Stage

    Parameters
    ----------
    serialnum : str
        Serial number of the stage.

    Returns
    -------
    kim_controller
        Thorlabs KIM Stage controller
    """
    kim_controller = importlib.import_module(
        "navigate.model.devices.APIs.thorlabs.kcube_inertial"
    )

    # Initialize
    kim_controller.TLI_BuildDeviceList()

    # Open the same serial number device if there are several devices connected to the
    # computer
    available_serialnum = kim_controller.TLI_GetDeviceListExt()
    if not list(filter(lambda s: str(s) == str(serialnum), available_serialnum)):
        print(
            f"** Please make sure Thorlabs stage with serial number {serialnum} "
            f"is connected to the computer!"
        )
        raise RuntimeError
    kim_controller.KIM_Open(str(serialnum))
    return kim_controller


class TLKIMStage(StageBase):
    """Thorlabs KIM Stage"""

    def __init__(self, microscope_name, device_connection, configuration, device_id=0):
        """Initialize the stage.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : str
            Connection string for the device.
        configuration : dict
            Configuration dictionary for the device.
        device_id : int
            Device ID for the device.
        """
        super().__init__(
            microscope_name, device_connection, configuration, device_id
        )  # only initialize the focus axis

        # Default mapping from self.axes to corresponding KIM channels
        axes_mapping = {"x": 4, "y": 2, "z": 3, "f": 1}
        if not self.axes_mapping:
            #: dict: Dictionary mapping software axes to hardware axes.
            self.axes_mapping = {
                axis: axes_mapping[axis] for axis in self.axes if axis in axes_mapping
            }

        #: list: List of KIM axes available.
        self.kim_axes = list(self.axes_mapping.values())

        if device_connection is not None:
            #: object: Thorlabs KIM Stage controller
            self.kim_controller = device_connection

        device_config = configuration["configuration"]["microscopes"][microscope_name][
            "stage"
        ]["hardware"]
        if type(device_config) == ListProxy:
            #: str: Serial number of the stage.
            self.serial_number = str(device_config[device_id]["serial_number"])
        else:
            self.serial_number = device_config["serial_number"]

    def __del__(self):
        """Delete the KIM Connection"""
        try:
            self.stop()
            self.kim_controller.KIM_Close(self.serial_number)
        except AttributeError:
            pass

    def report_position(self):
        """Report the position of the stage.

        Reports the position of the stage for all axes, and creates the hardware
        position dictionary.

        Returns
        -------
        position_dict : dict
            Dictionary containing the current position of the stage.
        """
        for ax, i in self.axes_mapping.items():
            try:
                # need to request before we get the current position
                self.kim_controller.KIM_RequestCurrentPosition(self.serial_number, i)
                pos = self.kim_controller.KIM_GetCurrentPosition(self.serial_number, i)
                setattr(self, f"{ax}_pos", pos)
            except (
                self.kim_controller.TLFTDICommunicationError,
                self.kim_controller.TLDLLError,
                self.kim_controller.TLMotorDLLError,
            ):
                pass

        return self.get_position_dict()

    def move_axis_absolute(self, axis, abs_pos, wait_until_done=False):
        """Implement movement logic along a single axis.

        Parameters
        ----------
        axis : str
            An axis. For example, 'x', 'y', 'z', 'f', 'theta'.
        abs_pos : float
            Absolute position value
        wait_until_done : bool
            Block until stage has moved to its new spot.

        Returns
        -------
        bool
            Was the move successful?
        """
        if axis not in self.axes_mapping:
            return False

        axis_abs = self.get_abs_position(axis, abs_pos)
        if axis_abs == -1e50:
            return False

        self.kim_controller.KIM_MoveAbsolute(
            self.serial_number, self.axes_mapping[axis], int(axis_abs)
        )

        if wait_until_done:
            stage_pos, n_tries, i = -1e50, 10, 0
            target_pos = axis_abs
            while (stage_pos != target_pos) and (i < n_tries):
                # TODO: do we need to request before we get the current position
                # self.kim_controller.KIM_RequestCurrentPosition(self.serial_number,
                # int(self.axes_mapping[axis]))
                stage_pos = self.kim_controller.KIM_GetCurrentPosition(
                    self.serial_number, int(self.axes_mapping[axis])
                )
                i += 1
                time.sleep(0.01)
            if stage_pos != target_pos:
                return False
        return True

    def move_absolute(self, move_dictionary, wait_until_done=False):
        """Move stage along a single axis.

        Parameters
        ----------
        move_dictionary : dict
            A dictionary of values required for movement. Includes 'x_abs', etc. for
            one or more axes. Expects values in micrometers, except for theta, which is
            in degrees.
        wait_until_done : bool
            Block until stage has moved to its new spot.

        Returns
        -------
        success : bool
            Was the move successful?
        """

        result = True
        for ax, n in self.axes_mapping.items():
            if f"{ax}_abs" not in move_dictionary:
                continue
            result = (
                self.move_axis_absolute(
                    ax, move_dictionary[f"{ax}_abs"], wait_until_done
                )
                and result
            )

        return result

    def stop(self):
        """Stop all stage channels move"""
        for i in self.kim_axes:
            self.kim_controller.KIM_MoveStop(self.serial_number, i)



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#
#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.
#
# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Standard Imports
import logging
import time

# Third Party Imports
from pipython import GCSDevice, pitools, GCSError

# Local Imports
from navigate.model.devices.stages.stage_base import StageBase

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


def build_PIStage_connection(controller_name, serial_number, stages, reference_modes):
    """Connect to the Physik Instrumente (PI) Stage

    Parameters
    ----------
    controller_name : str
        Name of the controller, e.g., "C-863.11"
    serial_number : str
        Serial number of the controller, e.g., "0112345678"
    stages : str
        Stages to connect to, e.g., "M-111.1DG"
    reference_modes : str
        Reference modes for the stages, e.g., "FRF"

    Returns
    -------
    stage_connection : dict
        Dictionary containing the pi_tools and pi_device objects
    """
    pi_stages = stages.split()
    pi_reference_modes = reference_modes.split()
    pi_tools = pitools
    pi_device = GCSDevice(controller_name)
    pi_device.ConnectUSB(serialnum=serial_number)
    pi_tools.startup(
        pi_device, stages=list(pi_stages), refmodes=list(pi_reference_modes)
    )

    # wait until pi_device is ready
    block_flag = True
    while block_flag:
        if pi_device.IsControllerReady():
            block_flag = False
        else:
            time.sleep(0.1)

    stage_connection = {"pi_tools": pi_tools, "pi_device": pi_device}
    return stage_connection


class PIStage(StageBase):
    """Physik Instrumente (PI) Stage Class

    Parameters
    ----------
    microscope_name : str
        Name of microscope in configuration
    device_connection : object
        Hardware device to connect to
    configuration : multiprocessing.managers.DictProxy
        Global configuration of the microscope

    """

    def __init__(self, microscope_name, device_connection, configuration, device_id=0):
        super().__init__(microscope_name, device_connection, configuration, device_id)

        # Default mapping from self.axes to corresponding PI axis labelling
        axes_mapping = {"x": 1, "y": 2, "z": 3, "f": 5, "theta": 4}

        if device_connection is not None:
            #: object: Physik Instrumente (PI) device
            self.pi_tools = device_connection["pi_tools"]

            #: object: Physik Instrumente (PI) device
            self.pi_device = device_connection["pi_device"]

            # Non-default axes_mapping
            axes_mapping = {x: y for x, y in zip(self.axes, self.pi_device.axes)}

        if not self.axes_mapping:
            self.axes_mapping = {
                axis: axes_mapping[axis] for axis in self.axes if axis in axes_mapping
            }

        #: list: List of PI axes available.
        self.pi_axes = list(self.axes_mapping.values())

    def __del__(self):
        """Delete the PI Connection



        Raises
        ------
        GCSError
            If the PI connection cannot be closed
        """
        try:
            if hasattr(self, "pi_device"):
                self.stop()
            logger.debug("PI connection closed")
        except (AttributeError, GCSError) as e:  # except BaseException:
            print("Error while disconnecting the PI stage")
            logger.exception(f"Error while disconnecting the PI stage - {e}")
            raise

    def report_position(self):
        """Reports the position for all axes, and create position dictionary.

        Positions from Physik Instrumente device are in millimeters

        Returns
        -------
        position_dict : dict
            Dictionary containing the position of all axes
        """
        for _ in range(10):
            try:
                positions = self.pi_device.qPOS(self.pi_axes)

                # convert to um
                for ax, n in self.axes_mapping.items():
                    pos = positions[str(n)]
                    if ax != "theta":
                        pos = round(pos * 1000, 2)
                    setattr(self, f"{ax}_pos", pos)
                break
            except GCSError as e:
                print("Physik Instrumente: Failed to report position")
                logger.exception(f"report_position failed - {e}")
                time.sleep(0.01)

        return self.get_position_dict()

    def move_axis_absolute(self, axis, abs_pos, wait_until_done=False):
        """Move stage along a single axis.

        Parameters
        ----------
        axis : str
            An axis. For example, 'x', 'y', 'z', 'f', 'theta'.
        abs_pos : float
            Absolute position value
        wait_until_done : bool
            Block until stage has moved to its new spot.

        Returns
        -------
        bool
            Was the move successful?
        """
        if axis not in self.axes_mapping:
            return False

        axis_abs = self.get_abs_position(axis, abs_pos)
        if axis_abs == -1e50:
            return False

        # Move the stage
        try:
            axis_num = self.axes_mapping[axis]
            pos = axis_abs
            if axis != "theta":
                pos /= 1000  # convert to mm
            self.pi_device.MOV({axis_num: pos})
        except GCSError as e:
            logger.exception(f"move_axis_absolute failed - {e}")
            return False

        if wait_until_done is True:
            return self.wait_on_target(axes=[axis_num])

        return True

    def move_absolute(self, move_dictionary, wait_until_done=False):
        """Move Absolute Method.

        XYZF Values are converted to millimeters for PI API.
        Theta Values are not converted.

        Parameters
        ----------
        move_dictionary : dict
            A dictionary of values required for movement. Includes 'x_abs', etc. for one
            or more axes. Expect values in micrometers, except for theta, which is
            in degrees.
        wait_until_done : bool
            Block until stage has moved to its new spot.

        Returns
        -------
        bool
            Was the move successful?
        """
        abs_pos_dict = self.verify_abs_position(move_dictionary)
        if not abs_pos_dict:
            return False

        pos_dict = {
            self.axes_mapping[axis]: abs_pos_dict[axis] / 1000
            if axis != "theta"
            else abs_pos_dict[axis]
            for axis in abs_pos_dict
        }

        try:
            self.pi_device.MOV(pos_dict)
        except GCSError as e:
            logger.exception(f"move_axis_absolute failed - {e}")
            return False

        if wait_until_done is True:
            return self.wait_on_target(axes=list(pos_dict.keys()))

        return True

    def stop(self):
        """Stop all stage movement abruptly."""
        try:
            self.pi_device.STP(noraise=True)
        except GCSError as e:
            logger.exception(f"Stage stop failed - {e}")

    def wait_on_target(self, axes=None):
        """Wait on target

        Parameters
        ----------
        axes : list or tuple or str
            Axes to wait for or None to wait for all axes.

        Returns
        -------
        bool
            Was the wait on target delay successful?
        """
        try:
            self.pi_tools.waitontarget(
                self.pi_device, axes=axes, timeout=1.75, polldelay=0.01
            )
        except GCSError as e:
            print("Wait on target failed")
            logger.exception(f"Wait on target failed - {e}")
            return False
        except Exception as e:
            logger.exception(f"Wait on target timeout error - {e}")
            return False
        return True
